import concurrent.futures
import logging
import pathlib
import re
import sys
import threading
from csv import QUOTE_NONE
from datetime import datetime, timedelta
from typing import Dict, Iterable, List, Optional, Tuple

import grpc
from diskcache import Cache
from ethereumetl.jobs.export_blocks_job import ExportBlocksJob
from ethereumetl.jobs.export_receipts_job import ExportReceiptsJob
from ethereumetl.jobs.export_traces_job import (
    ExportTracesJob as EthereumExportTracesJob,
)
from ethereumetl.providers.auto import get_provider_from_uri
from ethereumetl.service.eth_service import EthService
from ethereumetl.streaming.enrich import enrich_transactions
from ethereumetl.streaming.eth_item_id_calculator import EthItemIdCalculator
from ethereumetl.streaming.eth_item_timestamp_calculator import (
    EthItemTimestampCalculator,
)
from ethereumetl.thread_local_proxy import ThreadLocalProxy
from web3 import Web3

from ..config import GRAPHSENSE_DEFAULT_DATETIME_FORMAT, get_reorg_backoff_blocks
from ..datatypes import BadUserInputError
from ..db import AnalyticsDb
from ..utils import (
    batch,
    check_timestamp,
    first_or_default,
    hex_to_bytes,
    parse_timestamp,
    remove_prefix,
)
from ..utils.account import get_id_group
from ..utils.cache import TableBasedCache
from ..utils.logging import configure_logging, suppress_log_level
from ..utils.signals import graceful_ctlc_shutdown
from ..utils.tron import evm_to_bytes, strip_tron_prefix
from .common import (
    AbstractETLStrategy,
    AbstractTask,
    StoreTask,
    cassandra_ingest,
    write_to_sinks,
)
from .csv import (
    BLOCK_HEADER,
    LOGS_HEADER,
    TRACE_HEADER,
    TX_HEADER,
    format_blocks_csv,
    format_logs_csv,
    format_traces_csv,
    format_transactions_csv,
    write_csv,
)
from .tron.export_traces_job import TronExportTracesJob
from .tron.grpc.api.tron_api_pb2 import EmptyMessage, NumberMessage
from .tron.grpc.api.tron_api_pb2_grpc import WalletStub
from .tron.txTypeTransformer import TxTypeTransformer

logger = logging.getLogger(__name__)

BLOCK_BUCKET_SIZE = 1_000
TX_HASH_PREFIX_LEN = 5

PARQUET_PARTITION_SIZE = 100_000

WEB3_QUERY_BATCH_SIZE = 50
WEB3_QUERY_WORKERS = 40


class InMemoryItemExporter:
    """In-memory item exporter for EthStreamerAdapter export jobs."""

    def __init__(self, item_types: Iterable) -> None:
        self.item_types = item_types
        self.items: Dict[str, List] = {}

    def open(self) -> None:  # noqa
        """Open item exporter."""
        for item_type in self.item_types:
            self.items[item_type] = []

    def export_item(self, item) -> None:
        """Export single item."""
        item_type = item.get("type", None)
        if item_type is None:
            raise ValueError(f"type key is not found in item {item}")

        self.items[item_type].append(item)

    def close(self) -> None:
        """Close item exporter."""

    def get_items(self, item_type) -> Iterable:
        """Get items from exporter."""
        return self.items[item_type]


class AccountStreamerAdapter:
    """Standard Ethereum API style streaming adapter to export blocks, transactions,
    receipts, logs and traces."""

    # TODO: the adapter setup is sub optimal,
    # currently we create a Exporter*Job per call/batch
    # which in turn spins up a new thread pool for every call
    # which is not efficient and unnecessary overhead

    def __init__(
        self,
        batch_web3_provider: ThreadLocalProxy,
        batch_size: int = None,
        max_workers: int = None,
    ) -> None:
        self.batch_web3_provider = batch_web3_provider
        self.batch_size = batch_size
        self.max_workers = max_workers
        self.item_id_calculator = EthItemIdCalculator()
        self.item_timestamp_calculator = EthItemTimestampCalculator()

    def export_blocks_and_transactions(
        self,
        start_block: int,
        end_block: int,
        export_blocks: bool = True,
        export_transactions: bool = True,
    ) -> Tuple[Iterable, Iterable]:
        """Export blocks and transactions for specified block range."""

        blocks_and_transactions_item_exporter = InMemoryItemExporter(
            item_types=["block", "transaction"]
        )
        blocks_and_transactions_job = ExportBlocksJob(
            start_block=start_block,
            end_block=end_block,
            batch_size=self.batch_size,
            batch_web3_provider=self.batch_web3_provider,
            max_workers=self.max_workers,
            item_exporter=blocks_and_transactions_item_exporter,
            export_blocks=export_blocks,
            export_transactions=export_transactions,
        )

        blocks_and_transactions_job.run()
        blocks = blocks_and_transactions_item_exporter.get_items("block")
        transactions = blocks_and_transactions_item_exporter.get_items("transaction")
        return blocks, transactions

    def export_receipts_and_logs(
        self, transactions: Iterable
    ) -> Tuple[Iterable, Iterable]:
        """Export receipts and logs for specified transaction hashes."""

        exporter = InMemoryItemExporter(item_types=["receipt", "log"])
        job = ExportReceiptsJob(
            transaction_hashes_iterable=(
                transaction["hash"] for transaction in transactions
            ),
            batch_size=self.batch_size,
            batch_web3_provider=self.batch_web3_provider,
            max_workers=self.max_workers,
            item_exporter=exporter,
            export_receipts=True,
            export_logs=True,
        )

        job.run()
        receipts = exporter.get_items("receipt")
        logs = exporter.get_items("log")
        return receipts, logs

    def export_traces(
        self,
        start_block: int,
        end_block: int,
        include_genesis_traces: bool = True,
        include_daofork_traces: bool = False,
    ) -> Iterable[Dict]:
        """Export traces for specified block range."""

        exporter = InMemoryItemExporter(item_types=["trace"])
        job = EthereumExportTracesJob(
            start_block=start_block,
            end_block=end_block,
            batch_size=self.batch_size,
            web3=ThreadLocalProxy(lambda: Web3(self.batch_web3_provider)),
            max_workers=self.max_workers,
            item_exporter=exporter,
            include_genesis_traces=include_genesis_traces,
            include_daofork_traces=include_daofork_traces,
        )
        job.run()
        traces = exporter.get_items("trace")
        return traces, None


class EthStreamerAdapter(AccountStreamerAdapter):
    """Ethereum API style streaming adapter to export blocks, transactions,
    receipts, logs and traces."""


class TronStreamerAdapter(AccountStreamerAdapter):
    """Tron API style streaming adapter to export blocks, transactions,
    receipts, logs and traces.
    """

    def __init__(
        self,
        batch_web3_provider: ThreadLocalProxy,
        grpc_endpoint: str,
        batch_size: int = None,
        max_workers: int = None,
    ) -> None:
        super().__init__(batch_web3_provider, batch_size, max_workers)
        self.grpc_endpoint = grpc_endpoint

    def export_traces(
        self,
        start_block: int,
        end_block: int,
        include_genesis_traces: bool = True,
        include_daofork_traces: bool = False,
    ) -> Iterable[Dict]:
        """Export traces for specified block range."""

        # exporter = InMemoryItemExporter(item_types=["trace"])
        job = TronExportTracesJob(
            start_block=start_block,
            end_block=end_block,
            batch_size=self.batch_size,
            grpc_endpoint=self.grpc_endpoint,
            max_workers=self.max_workers,
        )

        return job.run()
        # traces = exporter.get_items("trace")
        # return traces

    def export_hash_to_type_mappings(
        self, transactions: Iterable, blocks: Iterable
    ) -> Dict:
        grpc_endpoint = remove_prefix(self.grpc_endpoint, "grpc://")
        channel = grpc.insecure_channel(grpc_endpoint)
        wallet_stub = WalletStub(channel)

        def get_type(tx):
            type_container = tx.raw_data.contract
            assert len(type_container) == 1
            return type_container[0].type

        def get_block(i):  # contains type
            msg = NumberMessage(num=i)
            info = wallet_stub.GetBlockByNum(msg)
            return info

        block_ids = [b["block_id"] for b in blocks]
        blocks_data = [get_block(b) for b in block_ids]
        txs_grpc = [tx for block in blocks_data for tx in block.transactions]
        types = [get_type(tx) for tx in txs_grpc]
        tx_hashes = [tx["hash"] for tx in transactions]
        hash_to_type = dict(zip(tx_hashes, types))

        # def getTransactionById(tx_hash):
        #    msg = BytesMessage(value=bytes.fromhex(tx_hash[2:]))
        #    info = wallet_stub.GetTransactionById(msg)
        #    return info
        # types2 = [get_type(getTransactionById(tx_hash)) for tx_hash in tx_hashes]
        # import numpy as np
        # assert np.all(np.array(types) == np.array(types2))

        return hash_to_type

    def get_trc10_token_infos(self) -> Iterable[Dict]:
        """Get all trc10 tokens from a grpc endpoint"""

        grpc_endpoint = remove_prefix(self.grpc_endpoint, "grpc://")

        channel = grpc.insecure_channel(grpc_endpoint)
        wallet_stub = WalletStub(channel)
        trc10_tokens = wallet_stub.GetAssetIssueList(EmptyMessage())

        list_of_dicts = []
        for token in trc10_tokens.assets:
            token_dict = {}
            for field in token.DESCRIPTOR.fields:
                value = getattr(token, field.name)
                token_dict[field.name] = value

            list_of_dicts.append(token_dict)

        return list_of_dicts


def get_last_block_yesterday(batch_web3_provider: ThreadLocalProxy) -> int:
    """Return last block number of previous day from Ethereum client."""

    web3 = Web3(batch_web3_provider)
    eth_service = EthService(web3)

    prev_date = datetime.date(datetime.today()) - timedelta(days=1)
    _, end_block = eth_service.get_block_range_for_date(prev_date)
    logger.info(
        f"Determining latest block before {prev_date.isoformat()} its {end_block:,}",
    )
    return end_block


def get_last_synced_block(batch_web3_provider: ThreadLocalProxy) -> int:
    """Return last synchronized block number from Ethereum client."""

    return int(Web3(batch_web3_provider).eth.getBlock("latest").number)


def ingest_configuration_cassandra(
    db: AnalyticsDb,
    block_bucket_size: int,
    tx_hash_prefix_len: int,
) -> None:
    """Store configuration details in Cassandra table."""
    cassandra_ingest(
        db,
        "configuration",
        [
            {
                "id": db.raw.keyspace_name(),
                "block_bucket_size": int(block_bucket_size),
                "tx_prefix_length": tx_hash_prefix_len,
            }
        ],
    )


def prepare_logs_inplace(items: Iterable, block_bucket_size: int):
    blob_colums = [
        "block_hash",
        "address",
        "data",
        "topic0",
        "tx_hash",
    ]
    for item in items:
        # remove column
        item.pop("type")
        # rename/add columns
        item["tx_hash"] = item.pop("transaction_hash")
        item["block_id"] = item.pop("block_number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)

        # Used for partitioning in parquet files
        # ignored otherwise
        item["partition"] = item["block_id"] // PARQUET_PARTITION_SIZE

        tpcs = item["topics"]

        if tpcs is None:
            tpcs = []

        if "topic0" not in item:
            # bugfix do not use None for topic0 but 0x, None
            # gets converted to UNSET which is not allowed for
            # key columns in cassandra and can not be filtered
            item["topic0"] = tpcs[0] if len(tpcs) > 0 else "0x"

        item["topics"] = [hex_to_bytes(t) for t in tpcs]

        # if topics contain duplicates
        if (
            len(item["topics"]) % 2 == 0 and len(item["topics"]) > 0
        ):  # todo may be removed if we are that there are no duplicates
            if (
                item["topics"][: len(item["topics"]) // 2]
                == item["topics"][len(item["topics"]) // 2 :]
            ):
                logger.warning(
                    f"duplicate found; hash: {item['tx_hash']};"
                    f" topics: {item['topics']}"
                )

        if "transaction_hash" in item:
            item.pop("transaction_hash")

        for elem in blob_colums:
            item[elem] = hex_to_bytes(item[elem])


def ingest_logs(
    items: Iterable,
    db: AnalyticsDb,
    sink_config: dict,
    block_bucket_size: int,
) -> None:
    """Ingest blocks into Apache Cassandra."""

    write_to_sinks(db, sink_config, "log", items)


def prepare_blocks_inplace_eth(
    items: Iterable,
    block_bucket_size: int,
):
    blob_colums = [
        "block_hash",
        "parent_hash",
        "nonce",
        "sha3_uncles",
        "logs_bloom",
        "transactions_root",
        "state_root",
        "receipts_root",
        "miner",
        "extra_data",
    ]
    for item in items:
        # remove column
        item.pop("type")
        # rename/add columns
        item["block_id"] = item.pop("number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)
        item["block_hash"] = item.pop("hash")

        # Used for partitioning in parquet files
        # ignored otherwise
        item["partition"] = item["block_id"] // PARQUET_PARTITION_SIZE

        # convert hex strings to byte arrays (blob in Cassandra)
        for elem in blob_colums:
            item[elem] = hex_to_bytes(item[elem])


def prepare_blocks_inplace_trx(items, block_bucket_size):
    prepare_blocks_inplace_eth(items, block_bucket_size)

    for b in items:
        # b["timestamp"] = b["timestamp"]
        check_timestamp(b["timestamp"])


def prepare_transactions_inplace_eth(
    items: Iterable, tx_hash_prefix_len: int, block_bucket_size: int
):
    blob_colums = [
        "tx_hash",
        "from_address",
        "to_address",
        "input",
        "block_hash",
        "receipt_contract_address",
        "receipt_root",
    ]
    for item in items:
        # remove column
        item.pop("type")
        # rename/add columns
        item["tx_hash"] = item.pop("hash")
        hash_slice = slice(2, 2 + tx_hash_prefix_len)
        item["tx_hash_prefix"] = item["tx_hash"][hash_slice]
        item["block_id"] = item.pop("block_number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)

        # Used for partitioning in parquet files
        # ignored otherwise
        item["partition"] = item["block_id"] // PARQUET_PARTITION_SIZE

        # convert hex strings to byte arrays (blob in Cassandra)
        for elem in blob_colums:
            item[elem] = hex_to_bytes(item[elem])


def prepare_transactions_inplace_trx(
    items: Iterable, tx_hash_prefix_len: int, block_bucket_size: int
):
    prepare_transactions_inplace_eth(items, tx_hash_prefix_len, block_bucket_size)

    for tx in items:
        type_transformer = TxTypeTransformer()
        tx = type_transformer.transform(tx)
        check_timestamp(tx["block_timestamp"])


def prepare_traces_inplace_eth(items: Iterable, block_bucket_size: int):
    blob_colums = ["tx_hash", "from_address", "to_address", "input", "output"]
    for item in items:
        # remove column
        item.pop("type")
        # rename/add columns
        item["tx_hash"] = item.pop("transaction_hash")
        item["block_id"] = item.pop("block_number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)

        # Used for partitioning in parquet files
        # ignored otherwise
        item["partition"] = item["block_id"] // PARQUET_PARTITION_SIZE

        item["trace_address"] = (
            ",".join(map(str, item["trace_address"]))
            if item["trace_address"] is not None
            else None
        )
        # convert hex strings to byte arrays (blob in Cassandra)
        for elem in blob_colums:
            item[elem] = hex_to_bytes(item[elem])


def prepare_traces_inplace_trx(items: Iterable, block_bucket_size: int):
    blob_colums = ["tx_hash", "caller_address", "transferto_address"]
    for item in items:
        # rename/add columns
        item["tx_hash"] = item.pop("transaction_hash")
        item["block_id"] = item.pop("block_number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)
        item["transferto_address"] = item.pop("transferTo_address")

        # Used for partitioning in parquet files
        # ignored otherwise
        item["partition"] = item["block_id"] // PARQUET_PARTITION_SIZE

        # item["trace_address"] = (
        #    ",".join(map(str, item["trace_address"]))
        #    if item["trace_address"] is not None
        #    else None
        # )
        # convert hex strings to byte arrays (blob in Cassandra)
        for elem in blob_colums:
            item[elem] = evm_to_bytes(item[elem])


def prepare_fees_inplace(fees: Iterable, tx_hash_prefix_len: int) -> None:
    blob_colums = ["tx_hash"]
    for item in fees:
        hash_slice = slice(2, 2 + tx_hash_prefix_len)
        item["tx_hash_prefix"] = item["tx_hash"][hash_slice]

        for elem in blob_colums:
            item[elem] = evm_to_bytes(item[elem])


def prepare_trc10_tokens_inplace(items: Iterable):
    def decode_bytes(bytes_):
        if bytes_ is None:
            return None
        # \xa0 is not a valid utf-8 character
        bytes_ = bytes_.replace(b"\xa0", b" ")
        try:
            return bytes_.decode()
        except Exception:
            return bytes_.decode(encoding="Latin1")

    fields_to_convert = ["name", "abbr", "description", "url"]
    for item in items:
        for field in fields_to_convert:
            item[field] = decode_bytes(item.get(field, None))

    times = ["start_time", "end_time"]
    for item in items:
        for field in times:
            if field in item:
                item[field] = item[field] // 1000
                # there are some known anomalous timestamps in the data.
                # We dont want to send too many warnings therefore omit check.
                # check_timestamp(item[field])

    # cast id to int
    for item in items:
        item["id"] = int(item["id"])

    # cast frozen_supply to list of tuples
    for item in items:
        if len(item["frozen_supply"]) > 0:
            item["frozen_supply"] = [
                (x.frozen_amount, x.frozen_days) for x in item["frozen_supply"]
            ]

    for item in items:
        # already in bytes, so strip tron prefix should do the trick
        item["owner_address"] = strip_tron_prefix(item["owner_address"])


def print_block_info(
    last_synced_block: int, last_ingested_block: Optional[int]
) -> None:
    """Display information about number of synced/ingested blocks."""

    logger.info(f"Last synced block: {last_synced_block:,}")
    if last_ingested_block is None:
        logger.info("Last ingested block: None")
    else:
        logger.info(f"Last ingested block: {last_ingested_block:,}")


def get_connection_from_url(provider_uri: str, provider_timeout=600):
    return ThreadLocalProxy(
        lambda: get_provider_from_uri(
            provider_uri, timeout=provider_timeout, batch=True
        )
    )


def ingest(
    db: AnalyticsDb,
    currency: str,
    sources: List[str],
    sink_config: dict,
    user_start_block: Optional[int],
    user_end_block: Optional[int],
    batch_size: int,
    info: bool,
    previous_day: bool,
    provider_timeout: int,
    mode: str,
):
    logger.info("Writing data sequentially")
    # make sure that only supported sinks are selected.
    if not all((x in ["cassandra", "parquet"]) for x in sink_config.keys()):
        raise BadUserInputError(
            "Unsupported sink selected, supported: cassandra,"
            f" parquet; got {list(sink_config.keys())}"
        )

    logger.info(f"Writing data to {list(sink_config.keys())}")

    http_provider_uri = first_or_default(sources, lambda x: x.startswith("http"))

    if http_provider_uri is None:
        raise BadUserInputError("No http provider (node url) is configured.")

    thread_proxy = get_connection_from_url(http_provider_uri, provider_timeout)
    last_synced_block = get_last_synced_block(thread_proxy)
    last_ingested_block = db.raw.get_highest_block()
    print_block_info(last_synced_block, last_ingested_block)

    if currency == "trx":
        if user_start_block == 0:
            user_start_block = 1
            logger.warning(
                "Start was set to 1 since genesis blocks "
                "don't have logs and cause issues."
            )

        grpc_provider_uri = first_or_default(sources, lambda x: x.startswith("grpc"))
        if grpc_provider_uri is None:
            raise BadUserInputError("No grpc provider (node url) is configured.")

        adapter = TronStreamerAdapter(
            thread_proxy,
            grpc_endpoint=grpc_provider_uri,
            batch_size=WEB3_QUERY_BATCH_SIZE,
            max_workers=WEB3_QUERY_WORKERS,
        )

        # ingest trc10 table
        token_infos = adapter.get_trc10_token_infos()
        prepare_trc10_tokens_inplace(token_infos)
        write_to_sinks(db, sink_config, "trc10", token_infos)

        prepare_transactions_inplace = prepare_transactions_inplace_trx
        prepare_blocks_inplace = prepare_blocks_inplace_trx
        prepare_traces_inplace = prepare_traces_inplace_trx

    elif currency == "eth":
        adapter = EthStreamerAdapter(
            thread_proxy,
            batch_size=WEB3_QUERY_BATCH_SIZE,
            max_workers=WEB3_QUERY_WORKERS,
        )
        prepare_transactions_inplace = prepare_transactions_inplace_eth
        prepare_blocks_inplace = prepare_blocks_inplace_eth
        prepare_traces_inplace = prepare_traces_inplace_eth
    else:
        raise NotImplementedError(f"Currency {currency} not implemented")

    start_block = 0
    if user_start_block is None:
        if last_ingested_block is not None:
            start_block = last_ingested_block + 1
    else:
        start_block = user_start_block

    end_block = last_synced_block - get_reorg_backoff_blocks(currency)
    if user_end_block is not None:
        end_block = user_end_block

    if previous_day:
        end_block = get_last_block_yesterday(thread_proxy)

    if start_block > end_block:
        print("No blocks to ingest")
        return

    time1 = datetime.now()
    count = 0

    # if info then only print block info and exit
    if info:
        logger.info(
            f"Would ingest block range "
            f"{start_block:,} - {end_block:,} ({end_block-start_block:,} blks) "
            f"into {list(sink_config.keys())} "
        )

        return

    logger.info(
        f"Ingesting block range "
        f"{start_block:,} - {end_block:,} ({end_block-start_block:,} blks) "
        f"into {list(sink_config.keys())} "
    )

    with graceful_ctlc_shutdown() as check_shutdown_initialized:
        for block_id in range(start_block, end_block + 1, batch_size):
            current_end_block = min(end_block, block_id + batch_size - 1)

            with suppress_log_level(logging.INFO):
                blocks, txs = adapter.export_blocks_and_transactions(
                    block_id, current_end_block
                )
                receipts, logs = adapter.export_receipts_and_logs(txs)
                traces, fees = adapter.export_traces(
                    block_id, current_end_block, True, True
                )
                enriched_txs = enrich_transactions(txs, receipts)

            # reformat and edit data
            prepare_logs_inplace(logs, BLOCK_BUCKET_SIZE)
            prepare_traces_inplace(traces, BLOCK_BUCKET_SIZE)
            prepare_transactions_inplace(
                enriched_txs, TX_HASH_PREFIX_LEN, BLOCK_BUCKET_SIZE
            )
            prepare_blocks_inplace(blocks, BLOCK_BUCKET_SIZE)
            if fees is not None:
                prepare_fees_inplace(fees, TX_HASH_PREFIX_LEN)
                write_to_sinks(db, sink_config, "fee", fees)

            # ingest into Cassandra
            write_to_sinks(db, sink_config, "log", logs)
            write_to_sinks(db, sink_config, "trace", traces)
            write_to_sinks(db, sink_config, "transaction", enriched_txs)
            write_to_sinks(db, sink_config, "block", blocks)

            count += batch_size

            last_block = blocks[-1]
            last_block_ts = last_block["timestamp"]

            last_block_date = parse_timestamp(last_block_ts)
            time2 = datetime.now()
            time_delta = (time2 - time1).total_seconds()
            logger.info(
                f"Last processed block: {current_end_block:,} "
                f"[{last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)}] "
                f"({count/time_delta:.1f} blks/s)"
            )
            time1 = time2
            count = 0

            if check_shutdown_initialized():
                break

    last_block_date = parse_timestamp(last_block_ts)
    logger.info(
        f"Processed block range "
        f"{start_block:,} - {end_block:,} "
        f" ({last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)})"
    )

    # store configuration details
    if "cassandra" in sink_config.keys():
        ingest_configuration_cassandra(
            db, int(BLOCK_BUCKET_SIZE), int(TX_HASH_PREFIX_LEN)
        )


class LoadLogsTask(AbstractTask):
    def run(self, ctx, data):
        txs = data
        receipts, logs = ctx.adapter.export_receipts_and_logs(txs)
        enriched_txs = ctx.strategy.enrich_transactions(txs, receipts)
        logst = ctx.strategy.transform_logs(logs, ctx.BLOCK_BUCKET_SIZE)
        txst = ctx.strategy.transform_transactions(
            enriched_txs, ctx.TX_HASH_PREFIX_LEN, ctx.BLOCK_BUCKET_SIZE
        )
        return [(StoreTask(), ("transaction", txst)), (StoreTask(), ("log", logst))]


class LoadLogsAndTypeTask(AbstractTask):
    def __init__(
        self, is_update_transactions_mode: bool = False, blocks: Iterable = None
    ):
        self.is_update_transactions_mode = is_update_transactions_mode
        self.blocks = blocks

    def run(self, ctx, data):
        txs = data
        receipts, logs = ctx.adapter.export_receipts_and_logs(txs)
        hash_to_type = ctx.adapter.export_hash_to_type_mappings(txs, self.blocks)
        enriched_txs = ctx.strategy.enrich_transactions(txs, receipts)
        enriched_txs = ctx.strategy.enrich_transactions_with_type(
            enriched_txs, hash_to_type
        )
        txst = ctx.strategy.transform_transactions(
            enriched_txs, ctx.TX_HASH_PREFIX_LEN, ctx.BLOCK_BUCKET_SIZE
        )
        if self.is_update_transactions_mode:
            return [(StoreTask(), ("transaction", txst))]

        logst = ctx.strategy.transform_logs(logs, ctx.BLOCK_BUCKET_SIZE)
        return [(StoreTask(), ("transaction", txst)), (StoreTask(), ("log", logst))]


class LoadBlockTask(AbstractTask):
    def run(self, ctx, data):
        start, end = data
        blocks, txs = ctx.adapter.export_blocks_and_transactions(start, end)
        blockst = ctx.strategy.transform_blocks(blocks, ctx.BLOCK_BUCKET_SIZE)
        return [
            (StoreTask(), ("block", blockst)),
            (LoadLogsTask(), txs),
        ]


class LoadBlockTaskTrx(AbstractTask):
    def __init__(self, is_update_transactions_mode: bool = False):
        self.is_update_transactions_mode = is_update_transactions_mode

    def run(self, ctx, data):
        start, end = data
        blocks, txs = ctx.adapter.export_blocks_and_transactions(start, end)
        blockst = ctx.strategy.transform_blocks(blocks, ctx.BLOCK_BUCKET_SIZE)
        return [
            (StoreTask(), ("block", blockst)),
            (LoadLogsAndTypeTask(self.is_update_transactions_mode, blocks), txs),
        ]


class LoadTracesTask(AbstractTask):
    def __init__(self, fees_only_mode: bool = False):
        self.fees_only_mode = fees_only_mode

    def run(self, ctx, data):
        start, end = data
        traces, fees = ctx.adapter.export_traces(start, end, True, True)
        tracest = ctx.strategy.transform_traces(traces, ctx.BLOCK_BUCKET_SIZE)
        fees = ctx.strategy.transform_fees(fees, ctx.TX_HASH_PREFIX_LEN)

        tasks = []

        if not self.fees_only_mode:
            tasks.append((StoreTask(), ("trace", tracest)))

        if fees is not None:
            tasks.append((StoreTask(), ("fee", fees)))

        return tasks


class LoadTrc10TokenInfoTask(AbstractTask):
    def run(self, ctx, data):
        token_infos = ctx.adapter.get_trc10_token_infos()
        token_infost = ctx.strategy.transform_trc10_token_infos(token_infos)
        return [(StoreTask(), ("trc10", token_infost))]


class EthETLStrategy(AbstractETLStrategy):
    def __init__(
        self,
        http_provider_uri: str,
        provider_timeout: int,
        is_trace_only_mode: bool,
        fees_only_mode: bool = False,
    ):
        self.http_provider_uri = http_provider_uri
        self.provider_timeout = provider_timeout
        self.is_trace_only_mode = is_trace_only_mode
        self.fees_only_mode = fees_only_mode

    def per_blockrange_tasks(self):
        if self.is_trace_only_mode:
            return [LoadTracesTask(self.fees_only_mode)]
        else:
            return [LoadBlockTask(), LoadTracesTask()]

    def transform_logs(self, logs, block_bucket_size: int):
        prepare_logs_inplace(logs, block_bucket_size)
        return logs

    def transform_fees(self, fees, tx_hash_prefix_len: int):
        if fees is not None:
            prepare_fees_inplace(fees, tx_hash_prefix_len)
        return fees

    def transform_transactions(
        self, enriched_txs, tx_hash_prefix_len: int, block_bucket_size: int
    ):
        prepare_transactions_inplace_eth(
            enriched_txs, tx_hash_prefix_len, block_bucket_size
        )
        return enriched_txs

    def enrich_transactions(self, txs, receipts):
        return enrich_transactions(txs, receipts)

    def transform_traces(self, traces, block_bucket_size: int):
        prepare_traces_inplace_eth(traces, block_bucket_size)
        return traces

    def transform_blocks(self, blocks, block_bucket_size: int):
        prepare_blocks_inplace_eth(blocks, block_bucket_size)
        return blocks

    def get_source_adapter(self):
        return EthStreamerAdapter(
            get_connection_from_url(self.http_provider_uri, self.provider_timeout),
            batch_size=WEB3_QUERY_BATCH_SIZE,
            max_workers=WEB3_QUERY_WORKERS,
        )


class TrxETLStrategy(EthETLStrategy):
    def __init__(
        self,
        http_provider_uri: str,
        provider_timeout: int,
        grpc_provider_uri: str,
        is_trace_only_mode: bool,
        is_update_transactions_mode: bool,
        batch_size: int = WEB3_QUERY_BATCH_SIZE,
        fees_only_mode: bool = False,
    ):
        super().__init__(
            http_provider_uri, provider_timeout, is_trace_only_mode, fees_only_mode
        )
        self.batch_size = batch_size
        self.grpc_provider_uri = grpc_provider_uri
        self.is_update_transactions_mode = is_update_transactions_mode

    def pre_processing_tasks(self):
        return [LoadTrc10TokenInfoTask()]

    def transform_transactions(
        self, enriched_txs, tx_hash_prefix_len: int, block_bucket_size: int
    ):
        prepare_transactions_inplace_trx(
            enriched_txs, tx_hash_prefix_len, block_bucket_size
        )
        return enriched_txs

    def enrich_transactions_with_type(self, enriched_txs, hash_to_type):
        for tx in enriched_txs:
            tx["transaction_type"] = hash_to_type[tx["hash"]]

        return enriched_txs

    def transform_blocks(self, blocks, block_bucket_size: int):
        prepare_blocks_inplace_trx(blocks, block_bucket_size)
        return blocks

    def transform_traces(self, traces, block_bucket_size: int):
        prepare_traces_inplace_trx(traces, block_bucket_size)
        return traces

    def transform_trc10_token_infos(self, token_infos):
        prepare_trc10_tokens_inplace(token_infos)
        return token_infos

    def get_source_adapter(self):
        return TronStreamerAdapter(
            get_connection_from_url(self.http_provider_uri, self.provider_timeout),
            grpc_endpoint=self.grpc_provider_uri,
            batch_size=self.batch_size,
            max_workers=self.batch_size,
        )

    def per_blockrange_tasks(self):
        if self.is_trace_only_mode:
            return [LoadTracesTask(self.fees_only_mode)]
        elif self.is_update_transactions_mode:
            return [LoadBlockTaskTrx(self.is_update_transactions_mode)]
        else:
            return [LoadBlockTaskTrx(), LoadTracesTask()]


def ingest_async(
    db: AnalyticsDb,
    currency: str,
    sources: List[str],
    sink_config: dict,
    user_start_block: Optional[int],
    user_end_block: Optional[int],
    batch_size_user: int,
    info: bool,
    previous_day: bool,
    provider_timeout: int,
    mode: str,
):
    logger.info("Writing data in parallel")

    interleave_batches = 2
    batch_size = (
        (batch_size_user // interleave_batches)
        if batch_size_user >= interleave_batches
        else batch_size_user
    )

    # make sure that only supported sinks are selected.
    if not all((x in ["cassandra", "parquet", "fs-cache"]) for x in sink_config.keys()):
        raise BadUserInputError(
            "Unsupported sink selected, supported: cassandra,"
            f" parquet; got {list(sink_config.keys())}"
        )

    fees_only_mode = mode == "account_fees_only"
    is_trace_only_mode = mode == "account_traces_only" or fees_only_mode
    is_update_transactions_mode = mode == "trx_update_transactions"

    http_provider_uri = first_or_default(sources, lambda x: x.startswith("http"))
    if http_provider_uri is None:
        raise BadUserInputError("No http provider (node url) is configured.")

    thread_proxy = get_connection_from_url(http_provider_uri, provider_timeout)
    last_synced_block = get_last_synced_block(thread_proxy)
    last_ingested_block = db.raw.get_highest_block()
    print_block_info(last_synced_block, last_ingested_block)

    if currency == "trx":
        if user_start_block == 0:
            user_start_block = 1
            logger.warning(
                "Start was set to 1 since genesis blocks "
                "don't have logs and cause issues."
            )

        grpc_provider_uri = first_or_default(sources, lambda x: x.startswith("grpc"))
        if grpc_provider_uri is None:
            raise BadUserInputError("No grpc provider (node url) is configured.")

        transform_strategy = TrxETLStrategy(
            http_provider_uri,
            provider_timeout,
            grpc_provider_uri,
            is_trace_only_mode,
            is_update_transactions_mode,
            batch_size=batch_size,
            fees_only_mode=fees_only_mode,
        )

    elif currency == "eth":
        transform_strategy = EthETLStrategy(
            http_provider_uri, provider_timeout, is_trace_only_mode
        )

    else:
        raise NotImplementedError(f"Currency {currency} not implemented")

    logger.info(f"Writing data to {list(sink_config.keys())}")

    start_block = 0
    if user_start_block is None:
        if last_ingested_block is not None:
            start_block = last_ingested_block + 1
    else:
        start_block = user_start_block

    end_block = last_synced_block - get_reorg_backoff_blocks(currency)
    if user_end_block is not None:
        end_block = user_end_block

    if previous_day:
        end_block = get_last_block_yesterday(thread_proxy)

    if start_block > end_block:
        print("No blocks to ingest")
        return

    # if info then only print block info and exit
    if info:
        logger.info(
            f"Would ingest block range "
            f"{start_block:,} - {end_block:,} ({end_block - start_block +1:,} blks) "
            f"into {list(sink_config.keys())} "
        )

        return

    logger.info(
        f"Ingesting block range "
        f"{start_block:,} - {end_block:,} ({end_block - start_block +1:,} blks) "
        f"into {list(sink_config.keys())} "
    )

    thread_context = threading.local()

    def initializer_worker(thrd_ctx, db, sink_config, strategy, loglevel):
        configure_logging(loglevel)
        new_db_conn = db.clone()
        new_db_conn.open()
        thrd_ctx.db = new_db_conn

        if "fs-cache" in sink_config:
            odirectory = sink_config["fs-cache"]["output_directory"]
            sink_config["fs-cache"]["cache"] = TableBasedCache(
                Cache(odirectory, eviction_policy="none")
            )
        thrd_ctx.adapter = strategy.get_source_adapter()
        thrd_ctx.strategy = strategy
        thrd_ctx.sink_config = sink_config
        thrd_ctx.TX_HASH_PREFIX_LEN = TX_HASH_PREFIX_LEN
        thrd_ctx.BLOCK_BUCKET_SIZE = BLOCK_BUCKET_SIZE

    def process_task(thrd_ctx, task, data):
        return task.run(thrd_ctx, data)

    def submit_tasks(ex, thrd_ctx, tasks, data=None):
        return [
            ex.submit(
                process_task,
                thrd_ctx,
                cmd,
                data,
            )
            for cmd in tasks
        ]

    with graceful_ctlc_shutdown() as check_shutdown_initialized:
        with concurrent.futures.ThreadPoolExecutor(
            initializer=initializer_worker,
            initargs=(
                thread_context,
                db,
                sink_config,
                transform_strategy,
                logger.getEffectiveLevel(),
            ),
            max_workers=4,  # we write at most 4 tables in parallel
        ) as ex:
            time1 = datetime.now()
            count = 0

            # Add preprocessing tasks
            tasks = submit_tasks(
                ex, thread_context, transform_strategy.pre_processing_tasks()
            )

            for block_ids in batch(
                range(start_block, end_block + 1, batch_size), n=interleave_batches
            ):
                # Execute tasks
                ranges = [(s, min(end_block, s + batch_size - 1)) for s in block_ids]
                current_end_block = ranges[-1][1]
                block_id = ranges[-1][0]

                blockrange_tasks = transform_strategy.per_blockrange_tasks()

                # Submit tasks per block range
                for s, e in ranges:
                    tasks += submit_tasks(
                        ex, thread_context, blockrange_tasks, data=(s, e)
                    )

                blocks = None
                while len(tasks) > 0:
                    completed = next(concurrent.futures.as_completed(tasks))
                    tasks.remove(completed)
                    for cont_task, data in completed.result():
                        # Get the data for the progress indicator
                        if isinstance(cont_task, StoreTask) and data[0] == "block":
                            blocks = data[1]

                        tasks += submit_tasks(
                            ex, thread_context, [cont_task], data=data
                        )

                # Update UI
                last_block_date_str = "Unknown"
                if not is_trace_only_mode:
                    last_block = blocks[-1]
                    last_block_ts = last_block["timestamp"]
                    last_blk_date = parse_timestamp(last_block_ts)
                    last_block_date_str = last_blk_date.strftime(
                        GRAPHSENSE_DEFAULT_DATETIME_FORMAT
                    )

                count += (current_end_block - block_id + 1) * interleave_batches

                # if count % 1000 == 0:
                time2 = datetime.now()
                time_delta = (time2 - time1).total_seconds()
                # logging.disable(logging.NOTSET)
                logger.info(
                    f"Last processed block: {current_end_block:,} "
                    f"[{last_block_date_str}] "  # noqa
                    f"({count/time_delta:.1f} blks/s)"
                )
                # logging.disable(logging.INFO)
                time1 = time2
                count = 0

                if check_shutdown_initialized():
                    break

    # logging.disable(logging.NOTSET)
    logger.info(
        f"Processed block range "
        f"{start_block:,} - {end_block:,} "
        f" ({last_block_date_str})"
    )

    # store configuration details
    if "cassandra" in sink_config.keys():
        ingest_configuration_cassandra(
            db, int(BLOCK_BUCKET_SIZE), int(TX_HASH_PREFIX_LEN)
        )


def export_csv(
    db: AnalyticsDb,
    currency: str,
    provider_uri: str,
    directory: str,
    user_start_block: Optional[int],
    user_end_block: Optional[int],
    continue_export: bool,
    batch_size: int,
    file_batch_size: int,
    partition_batch_size: int,
    info: bool,
    previous_day: bool,
    provider_timeout: int,
):
    logger.info(f"Writing data as csv to {directory}")

    thread_proxy = get_connection_from_url(provider_uri, provider_timeout)

    last_synced_block = get_last_synced_block(thread_proxy)
    last_ingested_block = db.raw.get_highest_block()
    print_block_info(last_synced_block, last_ingested_block)

    adapter = EthStreamerAdapter(
        thread_proxy, batch_size=WEB3_QUERY_BATCH_SIZE, max_workers=WEB3_QUERY_WORKERS
    )

    start_block = 0
    if user_start_block is None:
        if continue_export:
            block_files = sorted(pathlib.Path(directory).rglob("block*"))
            if block_files:
                last_file = block_files[-1].name
                logger.info(f"Last exported file: {block_files[-1]}")
                start_block = int(re.match(r".*-(\d+)", last_file).group(1)) + 1
    else:
        start_block = user_start_block

    end_block = get_last_synced_block(thread_proxy)
    logger.info(f"Last synced block: {end_block:,}")
    if user_end_block is not None:
        end_block = user_end_block
    if previous_day:
        end_block = get_last_block_yesterday(thread_proxy)

    time1 = datetime.now()
    count = 0

    block_bucket_size = file_batch_size
    if file_batch_size % batch_size != 0:
        logger.error("Error: file_batch_size is not a multiple of batch_size")
        sys.exit(1)

    if partition_batch_size % file_batch_size != 0:
        logger.error("Error: partition_batch_size is not a multiple of file_batch_size")
        sys.exit(1)

    rounded_start_block = start_block // block_bucket_size * block_bucket_size
    rounded_end_block = (end_block + 1) // block_bucket_size * block_bucket_size - 1

    if rounded_start_block > rounded_end_block:
        print("No blocks to export")
        return

    block_range = (
        rounded_start_block,
        rounded_start_block + block_bucket_size - 1,
    )

    if info:
        logger.info(
            f"Would process block range "
            f"{rounded_start_block:,} - {rounded_end_block:,}"
        )
        return

    path = pathlib.Path(directory)
    try:
        path.mkdir(parents=True, exist_ok=True)
    except (PermissionError, NotADirectoryError) as exception:
        logger.error(f"Could not output directory {directory}: {str(exception)}")
        sys.exit(1)

    block_file = "block_%08d-%08d.csv.gz" % block_range
    tx_file = "tx_%08d-%08d.csv.gz" % block_range
    trace_file = "trace_%08d-%08d.csv.gz" % block_range
    logs_file = "logs_%08d-%08d.csv.gz" % block_range

    logger.info(
        f"Processing block range " f"{rounded_start_block:,} - {rounded_end_block:,}"
    )

    block_list = []
    tx_list = []
    trace_list = []
    logs_list = []

    with graceful_ctlc_shutdown() as check_shutdown_initialized:
        for block_id in range(rounded_start_block, rounded_end_block + 1, batch_size):
            current_end_block = min(end_block, block_id + batch_size - 1)

            with suppress_log_level(logging.INFO):
                blocks, txs = adapter.export_blocks_and_transactions(
                    block_id, current_end_block
                )
                if (
                    block_id == 0 and currency == "trx"
                ):  # todo not necessary anymore, this case is impossible
                    # genesis tx of block 0 have no receipts
                    # to avoid errors we drop them
                    txs = [tx for tx in txs if tx["block_number"] > 0]
                receipts, logs = adapter.export_receipts_and_logs(txs)
                traces, fees = adapter.export_traces(
                    block_id, current_end_block, True, True
                )
                enriched_txs = enrich_transactions(txs, receipts)

            block_list.extend(format_blocks_csv(blocks))
            tx_list.extend(format_transactions_csv(enriched_txs, TX_HASH_PREFIX_LEN))
            trace_list.extend(format_traces_csv(traces))
            logs_list.extend(format_logs_csv(logs))

            count += batch_size

            last_block = block_list[-1]
            last_block_ts = last_block["timestamp"]
            last_block_date = parse_timestamp(last_block_ts)

            if count >= 1000:
                time2 = datetime.now()
                time_delta = (time2 - time1).total_seconds()
                logger.info(
                    f"Last processed block {current_end_block:,} "
                    f"[{last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)}] "
                    f"({count/time_delta:.1f} blks/s)"
                )
                time1 = time2
                count = 0

            if (block_id + batch_size) % block_bucket_size == 0:
                partition_start = block_id - (block_id % partition_batch_size)
                partition_end = partition_start + partition_batch_size - 1
                sub_dir = f"{partition_start:08d}-{partition_end:08d}"
                full_path = path / sub_dir
                full_path.mkdir(parents=True, exist_ok=True)

                if currency == "trx":
                    prepare_transactions_inplace_trx(tx_list)
                    prepare_blocks_inplace_trx(block_list)

                write_csv(full_path / trace_file, trace_list, TRACE_HEADER)

                write_csv(full_path / tx_file, tx_list, TX_HEADER)
                write_csv(full_path / block_file, block_list, BLOCK_HEADER)
                write_csv(
                    full_path / logs_file,
                    logs_list,
                    LOGS_HEADER,
                    delimiter="|",
                    quoting=QUOTE_NONE,
                )

                last_block = block_list[-1]
                last_block_ts = last_block["timestamp"]
                last_block_date = parse_timestamp(last_block_ts)

                logger.info(
                    f"Written blocks: {block_range[0]:,} - {block_range[1]:,} "
                    f"[{last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)}] "
                )

                block_range = (
                    block_id + batch_size,
                    block_id + batch_size + block_bucket_size - 1,
                )
                block_file = "block_%08d-%08d.csv.gz" % block_range
                tx_file = "tx_%08d-%08d.csv.gz" % block_range
                trace_file = "trace_%08d-%08d.csv.gz" % block_range
                logs_file = "logs_%08d-%08d.csv.gz" % block_range

                block_list.clear()
                tx_list.clear()
                trace_list.clear()
                logs_list.clear()

                if check_shutdown_initialized():
                    break

    logger.info(
        f"[{datetime.now()}] Processed block range "
        f"{rounded_start_block:,}:{rounded_end_block:,}"
        f" ({last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)})"
    )



from graphsenselib.utils.pipeline import Pipeline
from graphsenselib.utils.pipeline import Process as P
from graphsenselib.utils.pipeline import StopPipeline, set_context


def inc(ctx, item):
    return item + 1


def double(ctx, item):
    return item * 2


def add(ctx, item):
    return item + ctx.b


def raise_ex(ctx, item):
    raise ValueError("ohhh noooo")

    # s = [
    #     P(
    #         task=print_and_add_i,
    #         name="should be 11",
    #         init_func=set_context,
    #         init_args=({"b": 10},),
    #     )
    #     .pipe(
    #         P(
    #             task=print_and_add_i,
    #             name="should be 21",
    #             init_func=set_add,
    #             init_args=(10,),
    #         )
    #     )
    #     .pipe(
    #         P(
    #             task=print_and_add_i,
    #             name="should be 111",
    #             init_func=set_add,
    #             init_args=(100,),
    #         )
    #     ),
    #     P(task=print_and_add, name="should be 2").pipe(
    #         P(task=print_and_mult, name="should be 2")
    #     ),
    # ]


def test_simple_pipeline():
    s = [
        P(task=add, init_func=set_context, init_args=({"b": 10},))
        | P(task=add, init_func=set_context, init_args=({"b": 10},))
    ]

    with Pipeline(s) as p:
        p.put(1)
        p.put(100)
        p.put(StopPipeline())

        p.join()

        res = p.get_results()

    assert res == [21, 120]


def test_simple_pipeline_with_exeption1():
    s = [
        P(task=add, name="adder", init_func=set_context, init_args=({"b": 10},))
        | P(task=raise_ex, name="raiser")
    ]

    with Pipeline(s) as p:
        p.put(1)
        p.put(StopPipeline())

        p.join()

        res = p.get_results()

    assert len(res) == 1
    assert str(res[0].exc) == "ohhh noooo"
    assert type(res[0].exc) == ValueError


def test_simple_pipeline_with_exeption2():
    # exceptions are passed through and terminate workers
    s = [
        P(task=raise_ex, name="raiser")
        | P(task=add, name="adder", init_func=set_context, init_args=({"b": 10},))
    ]

    with Pipeline(s) as p:
        p.put(1)
        p.put(100)
        p.put(StopPipeline())
        p.put(StopPipeline())

        p.join()

        res = p.get_results()

    assert len(res) == 1
    assert str(res[0].exc) == "ohhh noooo"
    assert type(res[0].exc) == ValueError


def test_pipeline_with_downstream_exeption_and_no_stop():
    s = [
        P(task=add, name="adder", init_func=set_context, init_args=({"b": 10},))
        | P(task=raise_ex, name="raiser")
    ]

    with Pipeline(s) as p:
        p.put(1)
        # p.put(StopPipeline())

        p.join()

        res = p.get_results()

    assert len(res) == 1
    assert str(res[0].exc) == "ohhh noooo"
    assert type(res[0].exc) == ValueError


# def test_pipeline_with_downstream_exeption_and_no_stop():
#     s = [
#         P(task=add, name="adder", init_func=set_context, init_args=({"b": 10},))
#         | P(task=raise_ex, name="raiser")
#     ]

#     with Pipeline(s) as p:
#         p.put(1)
#         # p.put(StopPipeline())

#         p.join()

#         res = p.get_results()

#     assert len(res) == 1
#     assert str(res[0].exc) == "ohhh noooo"
#     assert type(res[0].exc) == ValueError



# -*- coding: utf-8 -*-
import json
import logging
from datetime import date, datetime
from typing import List, Optional

import pandas as pd
import requests

from graphsenselib.db import DbFactory
from graphsenselib.db.analytics import DATE_FORMAT
from graphsenselib.rates.coingecko import fetch_ecb_rates

logger = logging.getLogger(__name__)

MIN_START = "2010-07-17"


def cryptocompare_historical_url(start: str, end: str, symbol: str, fiat: str):
    # https://min-api.cryptocompare.com/documentation?key=Historical&cat=dataHistoday
    toDt = datetime.fromisoformat(end)
    sDt = datetime.fromisoformat(start)
    limit = (toDt - sDt).days + 1
    toTS = toDt.timestamp()
    # Docs of cryptocompre say max record limit is 2k so if more just load
    # the entire dataset
    if limit < 2000:
        return (
            "https://min-api.cryptocompare.com/data/v2/histoday"
            f"?fsym={symbol}&tsym={fiat}&toTS={toTS}&limit={limit}"
        )
    else:
        return (
            "https://min-api.cryptocompare.com/data/v2/histoday"
            f"?fsym={symbol}&tsym={fiat}&allData=true"
        )


def fetch_cryptocompare_rates(start: str, end: str, symbol: str, fiat: str):
    r1 = requests.get(cryptocompare_historical_url(start, end, symbol, fiat))
    rates = pd.DataFrame(json.loads(r1.content)["Data"]["Data"])
    rates["date"] = pd.to_datetime(rates["time"], unit="s").dt.floor("D")

    # assert rates.date[0] == pd.Timestamp(MIN_START)
    assert len(rates.date) == len(set(rates.date))
    rates["date_check"] = rates.date.diff()
    diffs = rates.date_check.value_counts()
    assert len(diffs) == 1
    assert diffs.keys().unique()[0] == pd.Timedelta("1 days")

    rates.date = rates.date.dt.strftime(DATE_FORMAT)
    rates.rename(columns={"close": fiat}, inplace=True)

    return rates[["date", fiat]]


def fetch(env, currency, fiat_currencies, start_date, end_date):
    with DbFactory().from_config(env, currency) as db:
        return fetch_impl(
            db,
            currency,
            fiat_currencies,
            start_date,
            end_date,
            None,
            start_date != MIN_START,
            False,
            True,
        )


def fetch_impl(
    db: Optional[object],
    currency: str,
    fiat_currencies: List[str],
    start_date: Optional[str],
    end_date: Optional[str],
    table: Optional[str],
    force: bool,
    dry_run: bool,
    abort_on_gaps: bool,
):
    if datetime.fromisoformat(start_date) < datetime.fromisoformat(MIN_START):
        start_date = MIN_START

        # query most recent data
    if not force and db:
        logger.info(f"Get last imported rate from {db.raw.get_keyspace()}")
        most_recent_date = db.raw.get_last_exchange_rate_date(table=table)
        if most_recent_date is not None:
            start_date = most_recent_date.strftime(DATE_FORMAT)

    logger.info(f"*** Fetch exchange rates for {currency} ***")
    logger.info(f"Start date: {start_date}")
    logger.info(f"End date: {end_date}")
    logger.info(f"Target fiat currencies: {fiat_currencies}")

    if datetime.fromisoformat(start_date) > datetime.fromisoformat(end_date):
        logger.error("Error: start date after end date.")
        raise SystemExit

    usd_rates = fetch_cryptocompare_rates(start_date, end_date, currency, "USD")

    ecb_rates = fetch_ecb_rates(fiat_currencies)

    # query conversion rates and merge converted values in exchange rates
    exchange_rates = usd_rates
    date_range = pd.date_range(
        date.fromisoformat(start_date), date.fromisoformat(end_date)
    )
    date_range = pd.DataFrame(date_range, columns=["date"])
    date_range = date_range["date"].dt.strftime("%Y-%m-%d")

    for fiat_currency in set(fiat_currencies) - {"USD"}:
        ecb_rate = ecb_rates[["date", fiat_currency]].rename(
            columns={fiat_currency: "fx_rate"}
        )
        merged_df = exchange_rates.merge(ecb_rate, on="date", how="left").merge(
            date_range, how="right"
        )

        # fill gaps over weekends
        merged_df["fx_rate"].fillna(method="ffill", inplace=True)
        merged_df["fx_rate"].fillna(method="bfill", inplace=True)

        if abort_on_gaps and merged_df["fx_rate"].isnull().values.any():
            logger.error(
                "Error: found missing values for currency "
                f"{fiat_currency}, aborting import. Probably a weekend."
            )
            logger.error(merged_df[merged_df["fx_rate"].isnull()])
            if not dry_run:
                # in case of dry run let it run
                # to see what would have been written to the db
                if len(merged_df[merged_df["fx_rate"].isnull()]) > 4:
                    # if missing more than 4 days, critical error
                    raise SystemExit(2)
                else:
                    raise SystemExit(15)
        merged_df[fiat_currency] = merged_df["USD"] * merged_df["fx_rate"]
        merged_df = merged_df[["date", fiat_currency]]
        exchange_rates = exchange_rates.merge(merged_df, on="date")

    return exchange_rates


def ingest(
    env,
    currency,
    fiat_currencies,
    start_date,
    end_date,
    table,
    force,
    dry_run,
    abort_on_gaps,
):
    if dry_run:
        logger.warning("This is a Dry-Run. Nothing will be written to the database!")
    with DbFactory().from_config(env, currency) as db:
        exchange_rates = fetch_impl(
            db,
            currency,
            fiat_currencies,
            start_date,
            end_date,
            table,
            force,
            dry_run,
            abort_on_gaps,
        )

        if exchange_rates.isna().values.any():
            logger.warning("exchange_rates contain NaNs, dropping them now")
            exchange_rates.dropna(inplace=True)

        # insert final exchange rates into Cassandra
        if "USD" not in fiat_currencies:
            exchange_rates.drop("USD", axis=1, inplace=True)
        exchange_rates["fiat_values"] = exchange_rates.drop("date", axis=1).to_dict(
            orient="records"
        )
        exchange_rates.drop(fiat_currencies, axis=1, inplace=True)

        # insert exchange rates into Cassandra table
        if not dry_run:
            logger.info(f"Writing to keyspace {db.raw.get_keyspace()}")
            if len(exchange_rates) > 0:
                db.raw.ingest(table, exchange_rates.to_dict("records"))
                logger.info(f"Inserted rates for {len(exchange_rates)} days: ")
                logger.info(
                    f"{exchange_rates.iloc[0].date} - {exchange_rates.iloc[-1].date}"
                )
            else:
                logger.info("Nothing to insert.")
        else:
            logger.info(
                "Dry run: No data inserted. "
                f"Would have inserted {len(exchange_rates)} days."
            )
            logger.info(exchange_rates)



from collections import namedtuple

from ..config import config, schema_types
from ..datatypes import (
    AddressAccount,
    AddressUtxo,
    TransactionHashAccount,
    TransactionHashUtxo,
)
from .account import RawDbAccount, RawDbAccountTrx, TransformedDbAccount
from .analytics import AnalyticsDb
from .analytics import KeyspaceConfig as KeyspaceConfigDB
from .cassandra import CassandraDb
from .utxo import RawDbUtxo, TransformedDbUtxo

DbTypeStrategy = namedtuple(
    "DatabaseStrategy",
    ["raw_db_type", "transformed_db_type", "address_type", "transaction_type"],
)


def get_db_types_by_schema_type(schema_type) -> DbTypeStrategy:
    if schema_type not in schema_types:
        raise ValueError(f"{schema_type} not yet defined.")

    if schema_type == "utxo":
        return DbTypeStrategy(
            RawDbUtxo, TransformedDbUtxo, AddressUtxo, TransactionHashUtxo
        )
    elif schema_type == "account":
        return DbTypeStrategy(
            RawDbAccount, TransformedDbAccount, AddressAccount, TransactionHashAccount
        )
    elif schema_type == "account_trx":
        return DbTypeStrategy(
            RawDbAccountTrx,
            TransformedDbAccount,
            AddressAccount,
            TransactionHashAccount,
        )
    else:
        raise ValueError(f"{schema_type} not yet supported.")


class DbFactory:
    def from_config(self, env, currency) -> AnalyticsDb:
        e = config.get_environment(env)
        ks = e.get_keyspace(currency)
        return self.from_name(
            ks.raw_keyspace_name,
            ks.transformed_keyspace_name,
            ks.schema_type,
            e.cassandra_nodes,
            currency,
        )

    def from_name(
        self,
        raw_keyspace_name,
        transformed_keyspace_name,
        schema_type,
        cassandra_nodes,
        currency,
    ) -> AnalyticsDb:
        db_types = get_db_types_by_schema_type(schema_type)
        return AnalyticsDb(
            raw=KeyspaceConfigDB(
                raw_keyspace_name,
                db_types.raw_db_type,
                db_types.address_type,
                db_types.transaction_type,
                currency,
            ),
            transformed=KeyspaceConfigDB(
                transformed_keyspace_name,
                db_types.transformed_db_type,
                db_types.address_type,
                db_types.transaction_type,
                currency,
            ),
            db=CassandraDb(cassandra_nodes),
        )



from typing import Iterable, Optional

from ..utils import flatten, hex_to_bytes
from ..utils.utxo import SlimTx, get_slim_tx_from_transaction
from .analytics import RawDb, TransformedDb


class TransformedDbUtxo(TransformedDb):
    def get_highest_cluster_id(self, sanity_check=True) -> Optional[int]:
        """Return last ingested cluster ID from cluster table."""
        ha = self._get_hightest_id(table="cluster", sanity_check=sanity_check)
        return ha


class RawDbUtxo(RawDb):
    def get_transaction_ids_in_block(self, block: int) -> Iterable:
        block_bucket_size = self.get_block_bucket_size()
        group = self.get_id_group(block, block_bucket_size)
        result = self.select(
            "block_transactions",
            columns=["txs"],
            where={"block_id_group": group, "block_id": block},
        )
        txs = flatten([block.txs for block in result])
        ids = [tx.tx_id for tx in txs]
        return ids

    def get_transactions_in_block(self, block: int) -> Iterable:
        tx_bucket_size = self.get_tx_bucket_size()
        minb = self.get_latest_tx_id_before_block(block)
        maxb = self.get_latest_tx_id_before_block(block + 1)

        mbg = self.get_id_group(minb, tx_bucket_size)
        mxbg = self.get_id_group(maxb, tx_bucket_size)

        rg = list(range(mbg, mxbg + 1))

        stmt = (
            f"select * from {self.get_keyspace()}.transaction where "
            "tx_id_group in :txidgroups and tx_id > :txid_lower "
            "and tx_id <= :txid_upper"
        )

        prepared_statement = self._db.get_prepared_statement(stmt)

        bstmt = prepared_statement.bind(
            {"txidgroups": rg, "txid_lower": minb, "txid_upper": maxb}
        )

        results = self._db.execute_statement(bstmt)

        return list(results)

    def get_addresses_in_block(self, block: int) -> Iterable[SlimTx]:
        tx_ids = self.get_transaction_ids_in_block(block)
        tx_bucket_size = self.get_tx_bucket_size()
        addresses = []
        for tx_id in tx_ids:
            group = self.get_id_group(tx_id, tx_bucket_size)
            result = self.select_one(
                "transaction",
                columns=["inputs", "outputs", "block_id", "tx_hash", "timestamp"],
                where={"tx_id_group": group, "tx_id": tx_id},
            )

            addresses += get_slim_tx_from_transaction(result)
        return addresses

    def get_tx_outputs(
        self,
        tx_hash: str,
        tx_prefix_length: Optional[int],
        tx_bucket_size: Optional[int],
    ) -> Optional[Iterable]:
        tx_prefix_length = tx_prefix_length or self.get_tx_prefix_length()
        tx_bucket_size = tx_bucket_size or self.get_tx_bucket_size()

        tx_id_record = self.select_one_safe(
            "transaction_by_tx_prefix",
            columns=["tx_id"],
            where={
                "tx_prefix": f"{tx_hash[:tx_prefix_length]}",
                "tx_hash": hex_to_bytes(tx_hash),
            },
        )
        if tx_id_record:
            tx_id = tx_id_record.tx_id
            result = self.select_one_safe(
                "transaction",
                columns=["outputs"],
                where={
                    "tx_id_group": self.get_id_group(tx_id, tx_bucket_size),
                    "tx_id": tx_id,
                },
            )

            res = {}
            for i, item in enumerate(result.outputs):
                res[i] = {
                    "addresses": item.address,
                    "value": item.value,
                    "type": item.address_type,
                }
            return res
        else:
            return None

    def get_latest_tx_id_before_block(self, block_id: int) -> Optional[int]:
        last_block = block_id - 1
        bucket_size = self.get_block_bucket_size()

        block = self.select_one_safe(
            "block_transactions",
            where={
                "block_id_group": self.get_id_group(last_block, bucket_size),
                "block_id": last_block,
            },
        )
        latest_tx_id = -1

        if not block:
            return latest_tx_id

        return max(tx.tx_id for tx in block.txs)



import gzip
from csv import DictWriter
from typing import Iterable

from ..utils.account import get_id_group

BLOCK_HEADER = [
    "parent_hash",
    "nonce",
    "sha3_uncles",
    "logs_bloom",
    "transactions_root",
    "state_root",
    "receipts_root",
    "miner",
    "difficulty",
    "total_difficulty",
    "size",
    "extra_data",
    "gas_limit",
    "gas_used",
    "timestamp",
    "transaction_count",
    "base_fee_per_gas",
    "block_id",
    "block_id_group",
    "block_hash",
]

TX_HEADER = [
    "nonce",
    "transaction_index",
    "from_address",
    "to_address",
    "value",
    "gas",
    "gas_price",
    "input",
    "block_timestamp",
    "block_hash",
    "max_fee_per_gas",
    "max_priority_fee_per_gas",
    "transaction_type",
    "receipt_cumulative_gas_used",
    "receipt_gas_used",
    "receipt_contract_address",
    "receipt_root",
    "receipt_status",
    "receipt_effective_gas_price",
    "tx_hash",
    "tx_hash_prefix",
    "block_id",
]

TRACE_HEADER = [
    "transaction_index",
    "from_address",
    "to_address",
    "value",
    "input",
    "output",
    "trace_type",
    "call_type",
    "reward_type",
    "gas",
    "gas_used",
    "subtraces",
    "trace_address",
    "error",
    "status",
    "trace_id",
    "trace_index",
    "tx_hash",
    "block_id",
    "block_id_group",
]

LOGS_HEADER = [
    "block_id_group",
    "block_id",
    "block_hash",
    "address",
    "data",
    "topics",
    "topic0",
    "tx_hash",
    "log_index",
    "transaction_index",
]


def format_blocks_csv(
    items: Iterable,
    block_bucket_size: int = 1_000,
) -> None:
    """Format blocks."""

    for item in items:
        # remove column
        item.pop("type")
        item.pop("withdrawals")
        item.pop("withdrawals_root")
        # rename/add columns
        item["block_id"] = item.pop("number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)
        item["block_hash"] = item.pop("hash")

    return items


def format_transactions_csv(
    items: Iterable,
    tx_hash_prefix_len: int = 4,
) -> None:
    """Format transactions."""

    for item in items:
        # remove column
        item.pop("type")
        item.pop("receipt_l1_fee")
        item.pop("receipt_l1_fee_scalar")
        item.pop("receipt_l1_gas_price")
        item.pop("receipt_l1_gas_used")

        # rename/add columns
        item["tx_hash"] = item.pop("hash")
        hash_slice = slice(2, 2 + tx_hash_prefix_len)
        item["tx_hash_prefix"] = item["tx_hash"][hash_slice]
        item["block_id"] = item.pop("block_number")

    return items


def format_traces_csv(
    items: Iterable,
    block_bucket_size: int = 1_000,
) -> None:
    """Format traces."""

    for item in items:
        # remove column
        item.pop("type")
        # rename/add columns
        item["tx_hash"] = item.pop("transaction_hash")
        item["block_id"] = item.pop("block_number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)
        item["trace_address"] = (
            "|".join(map(str, item["trace_address"]))
            if item["trace_address"] is not None
            else None
        )

    return items


def format_logs_csv(
    items: Iterable,
    block_bucket_size: int = 1_000,
) -> None:
    """Format logs."""

    for item in items:
        # remove column
        item.pop("type")
        # rename/add columns
        item["tx_hash"] = item.pop("transaction_hash")
        item["block_id"] = item.pop("block_number")
        item["block_id_group"] = get_id_group(item["block_id"], block_bucket_size)

        tpcs = item["topics"]

        if tpcs is None:
            tpcs = []

        if "topic0" not in item:
            item["topic0"] = tpcs[0] if len(tpcs) > 0 else None

        qt = ",".join([f'"{t}"' for t in tpcs])

        item["topics"] = f"[{qt}]"

        if "transaction_hash" in item:
            item.pop("transaction_hash")

    return items


def write_csv(
    filename: str,
    data: Iterable,
    header: Iterable,
    delimiter: str = ",",
    quoting=None,
    extrasaction="ignore",
) -> None:
    """Write list of dicts to compresses CSV file."""

    with gzip.open(filename, "wt") as csv_file:
        if quoting is None:
            csv_writer = DictWriter(
                csv_file,
                delimiter=delimiter,
                extrasaction=extrasaction,
                fieldnames=header,
            )
        else:
            csv_writer = DictWriter(
                csv_file,
                delimiter=delimiter,
                fieldnames=header,
                quoting=quoting,
                extrasaction=extrasaction,
                quotechar="",
            )
        csv_writer.writeheader()
        for row in data:
            csv_writer.writerow(row)



import logging
import threading
import time
from contextlib import contextmanager

import click
from rich.console import Console
from rich.logging import RichHandler

from ..config import GRAPHSENSE_DEFAULT_DATETIME_FORMAT

# create logger
logger = logging.getLogger(__name__)


@contextmanager
def suppress_log_level(loglevel: int):
    logging.disable(loglevel)
    try:
        yield
    finally:
        logging.disable(logging.NOTSET)


def configure_logging(loglevel):
    log_format = " | %(message)s"

    if loglevel < 10:
        # this means the value passed is
        # not a valid log level in python
        if loglevel == 0:
            loglevel = logging.WARNING
        elif loglevel == 1:
            loglevel = logging.INFO
        elif loglevel >= 2:
            loglevel = logging.DEBUG
            log_format = " | %(name)s | %(thread)d | %(message)s"

    """ RichHandler colorizes the logs """
    c = Console(width=220)
    if c.is_terminal:
        rh = RichHandler(rich_tracebacks=True, tracebacks_suppress=[click])
    else:
        # if file redirect set terminal width to 220
        rh = RichHandler(
            rich_tracebacks=True,
            tracebacks_suppress=[click],
            console=c,
            show_path=False,
        )
    logging.basicConfig(
        format=log_format,
        level=loglevel,
        datefmt=GRAPHSENSE_DEFAULT_DATETIME_FORMAT,
        handlers=[rh],
    )

    if loglevel <= logging.DEBUG:
        logger.debug("Logging set to verbose mode.")
        logging.getLogger("cassandra").setLevel(logging.DEBUG)
        logging.getLogger("ethereumetl").setLevel(logging.WARNING)
        logging.getLogger("web3").setLevel(logging.WARNING)
        logging.getLogger("Cluster").setLevel(logging.DEBUG)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.ERROR)
        logging.getLogger("ProgressLogger").setLevel(logging.ERROR)
        logging.getLogger("BatchWorkExecutor").setLevel(logging.ERROR)
    else:
        logging.getLogger("cassandra").setLevel(logging.ERROR)
        logging.getLogger("ethereumetl").setLevel(logging.WARNING)
        logging.getLogger("Cluster").setLevel(logging.ERROR)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.ERROR)
        logging.getLogger("ProgressLogger").setLevel(logging.ERROR)
        logging.getLogger("BatchWorkExecutor").setLevel(logging.ERROR)


class IndentLogger(logging.LoggerAdapter):
    _ident = 0

    def __init__(self, logger):
        super().__init__(logger, extra={})

    def set_ident(self, level):
        self._ident = level
        return self

    def process(self, msg, kwargs):
        return "{i}{m}".format(i=" " * self._ident, m=msg), kwargs


class LoggerScope:
    _active_group = threading.local()

    @classmethod
    def debug(Cls, logger, msg=None):
        return Cls(logger, msg=msg, level=logging.DEBUG)

    @classmethod
    def info(Cls, logger, msg=None):
        return Cls(logger, msg=msg, level=logging.INFO)

    @classmethod
    def error(Cls, logger, msg=None):
        return Cls(logger, msg=msg, level=logging.CRITICAL)

    @staticmethod
    def get_stack():
        if not hasattr(LoggerScope._active_group, "current"):
            LoggerScope._active_group.current = []
        return LoggerScope._active_group.current

    @staticmethod
    def get_current_scope():
        stk = LoggerScope.get_stack()
        return stk[0] if len(stk) > 0 else None

    @staticmethod
    def get_indent_logger(logger):
        return IndentLogger(logger).set_ident(len(LoggerScope.get_stack()) + 1)

    def __init__(self, logger, msg=None, level=logging.INFO):
        self._logger = logger
        self._level = level
        self._msg = msg

    def __enter__(self):
        currentstack = self.get_stack()
        self._start_time = time.time()
        if self._msg is not None:
            IndentLogger(self._logger).set_ident(len(currentstack)).log(
                self._level, f"B - {self._msg}"
            )
        currentstack.append(self)
        logger = IndentLogger(self._logger)
        logger.set_ident(len(currentstack) + 1)
        return logger

    def __exit__(self, exc_type, exc_value, exc_traceback):
        last = LoggerScope._active_group.current.pop()
        currentstacklen = len(self.get_stack())
        assert last == self, "Logger context exited out of order"
        self.elapsed_seconds = time.time() - self._start_time
        if self._msg is not None:
            IndentLogger(self._logger).set_ident(currentstacklen).log(
                self._level, f"E - {self._msg} - took {self.elapsed_seconds:.3f}s"
            )



import logging

logger = logging.getLogger(__name__)


class TxTypeTransformer:
    def __init__(self):
        self.type_to_transform_fkt = {
            0: self.AccountCreateContract,
            1: self.TransferContract,
            2: self.TransferAssetContract,
            3: self.no_handling,
            4: self.VoteWitnessContract,
            5: self.no_handling,
            6: self.no_handling,
            8: self.no_handling,
            9: self.no_handling,
            10: self.no_handling,
            11: self.FreezeBalanceContract,
            12: self.UnfreezeBalanceContract,
            13: self.WithdrawBalanceContract,
            14: self.no_handling,
            15: self.no_handling,
            16: self.no_handling,
            17: self.no_handling,
            18: self.no_handling,
            19: self.no_handling,
            20: self.no_handling,
            30: self.CreateSmartContract,
            31: self.TriggerSmartContract,
            32: self.no_handling,
            33: self.no_handling,
            41: self.no_handling,
            42: self.no_handling,
            43: self.no_handling,
            44: self.ExchangeTransactionContract,
            45: self.no_handling,
            46: self.AccountPermissionUpdateContract,
            48: self.no_handling,
            49: self.no_handling,
            51: self.no_handling,
            52: self.no_handling,
            53: self.no_handling,
            54: self.FreezeBalanceV2Contract,
            55: self.UnfreezeBalanceV2Contract,
            56: self.WithdrawExpireUnfreezeContract,
            57: self.DelegateResourceContract,
            58: self.UnDelegateResourceContract,
            59: self.CancelAllUnfreezeV2Contract,
        }

    def type_not_supported(self, type_):
        logger.warning(
            f"transaction_type {type_} not considered. Probably new. Check "
            "https://github.com/tronprotocol/java-tron/"
            "blob/develop/Tron%20protobuf%20protocol%20document.md"
            "and update the function mapping accordingly if necessary "
            "with a new transformation fkt."
        )
        return None

    def transform(self, x):
        type_ = x["transaction_type"]
        f = self.type_to_transform_fkt.get(type_)
        if f is None:
            self.type_not_supported(type_)
            return self.no_handling(x)
        return f(x)

    def no_handling(self, x):
        return x

    def WithdrawBalanceContract(self, x):
        """
        Example: e4f4ab696d4f3e00cfc41a7b89a93bd4b95b82d5faeaa70b9813c8aa558081da
        Swap to and from because tron shows that the recipient of the rewards is
        the sender of the transaction
        and sends those rewards to the null address. This does not represent the
        flow of funds, therefore we reverse it.
        """
        x["to_address"], x["from_address"] = x["from_address"], x["to_address"]
        return x

    def AccountCreateContract(self, x):
        """
        Example: a15559a627a9691097c6809be8d2815b768e41e045afeceaedaeda15c168c39c
        Value = 0 in test sample of 100 tx
        grpc_GetTransactionInfoById.fee = 1000000
        Fine
        """
        return x

    def TransferContract(self, x):
        """
        Example: dcdfb509b33d493e9553b279675a8e9053701cfc3088519f205feca302a3ff02
        Fine
        """
        return x

    def TransferAssetContract(self, x):
        """
        Example: 99cd1c4b193d2dd0ae703c457726c0dfe82ba2c785d8cad66f3b5c45fef25ec6
        Value = Number (In smallest denomination) of TRC10 Token transferred!
        This would break our pipeline because value is TRX for us, not some token
        """
        x["value"] = 0
        return x

    def VoteWitnessContract(self, x):
        """
        Example: 7f973ec59417fb12375b3475e16a4a4bb150cf0a8f64cb6e19093183525b215c
        Value = Number of Votes; not TRX -> set value to 0
        """
        x["value"] = 0
        return x

    def FreezeBalanceContract(self, x):
        """
        Example: Not found in the sample but to be safe:
        Set value to 0 like in UnfreezeBalanceContract
        """
        x["value"] = 0
        return x

    def UnfreezeBalanceContract(self, x):
        """
        Example: d2fcbf5cdeb3efc06d90012838c6e559d12dde1cf4711afe1e798e694c70595c
        from: unstaker
        to: None
        value: Quantity unstaked != 0
        """
        x["value"] = 0
        return x

    def CreateSmartContract(self, x):
        """
        Example: 140e3404392a690e31c7009b7cee2b8dc84ecf58c391cebbd531adf7772efae7
        from: Creator
        to: None
        Value: 0

        todo: In the small simulation, those tx appear duplicated very often.
            Might be worth it to check that in productive system aswell.
        fine
        """
        return x

    def TriggerSmartContract(self, x):
        """
        Example: a53187fb4b99c53e1447637285bc336679a5ce3d11857793391875ccbae214ca
        Example2: 8c9656fcc6bd588d5a02f5aaeb795ebcf6422919a353c0df36378927fe611027
        from: EOA
        to: CA
        value: Mostly 0, sometimes TRX value if TRX is paid

        Fine
        """
        return x

    def ExchangeTransactionContract(self, x):
        """
        Example: 49b2b69dc736323b99177b79c3c0e4945747cd5900ef7770bc0480b49dc73ae0
        from: owner
        to: None
        value: Quantity of asset consumed to buy TRX TODO: how much TRX gotten?
        todo: In the small simulation, those tx appear duplicated very often.
            Might be worth it to check that in productive system aswell.
        """
        return x

    def AccountPermissionUpdateContract(self, x):
        """
        Example: b820ae2b84983aa2a11bd4b4927b612b64f9a6064026275dcc75850398927044
        Update account permissions
        from: account owner
        to: None
        value: 100.000.000 (100 TRX)
        value == grpc_GetTransactionInfoById.fee
        """
        x["value"] = 0
        return x

    def FreezeBalanceV2Contract(self, x):
        """
        Example: b1509e72d6ac99e0e1efccfb91ddc0d1c0a6d4408c1810ce73fc9407692502b4
        sender account staked xxx TRX and obtained Energy & TRON Power via Stake 2.0
        from: staker
        to: None
        value: 0
        """

        return x

    def UnfreezeBalanceV2Contract(self, x):
        """
        Example: bb3e61ac76603e5a0043a6c3cfb3b702213f3a24da8f5ac6f6e6892bf317022c
        Unstake xxx asset in Stake2.0, Energy & Tron power are deducted from the account
        from: unstaker
        to: None
        value = 0 != Quantity of assets unstaked
        """
        return x

    def WithdrawExpireUnfreezeContract(self, x):
        """
        Example: a4e6b5521a5d7c68865a8d5e5eb7f181442adc4fbc6979f34711924c70f55346
        Withdraw unstaked asset
        Seems to be general (asset; not just TRX) but in the few contracts
        I checked it was only TRX
        from = Unstaker
        to = None
        Value = Quantity (smallest denomination) of that asset.
        """
        x["value"] = 0
        return x

    def DelegateResourceContract(self, x):
        """
        Example: edba09c24eabd3ddbcd0df53732c780be3f2a8e396babdf6b2e2c46c69d0a07e
        Value = 0 in test sample of 100 tx
        """
        return x

    def UnDelegateResourceContract(self, x):
        """
        Example: 0b51f06e4677125dedb753d1aee6be8942b4689377b95fb9ec12e0020b38a790
        Staked assets are released
        Value = 0;
        Dont yet understand "from" and "to" relationship, both seem to be
        regular accounts but with tons of tx
        """
        return x

    def CancelAllUnfreezeV2Contract(self, x):
        """
        Example: 45d69c16dcb2e18f59e9d1a71a6644456aef74c1460bae29dc7506b153f1ae3d
        Cancel Unstake
        from: unstaker canceller
        to: None
        Value: Tron value to cancel from unstaking != 0
        todo: In the small simulation, those tx appear duplicated very often.
            Might be worth it to check that in productive system aswell.
        """
        x["value"] = 0
        return x



from graphsenselib.datatypes.address import AddressUtxo
from graphsenselib.utils import DataObject as D


def test_is_bech32(capsys):
    a = AddressUtxo(
        "bc1qu5ujlp9dkvtgl98jakvw9ggj9uwyk79qhvwvrg",
        D(**{"bech_32_prefix": "bc", "address_prefix_length": 5}),
    )

    assert a.is_bech32 is True
    assert a.prefix == "1qu5u"


def test_isnot_bech32_if_no_prefix_config(capsys):
    a = AddressUtxo(
        "bc1qu5ujlp9dkvtgl98jakvw9ggj9uwyk79qhvwvrg",
        D(**{"bech_32_prefix": "", "address_prefix_length": 5}),
    )

    assert a.is_bech32 is False
    assert a.prefix == "bc1qu"


def test_isnot_bech32(capsys):
    a = AddressUtxo(
        "3MzhkeE3EjenfoaYm6KxLaDLryQ3ZYytF9",
        D(**{"bech_32_prefix": "bc", "address_prefix_length": 5}),
    )

    assert a.is_bech32 is False
    assert a.prefix == "3Mzhk"



# flake8: noqa: W291

from graphsenselib.schema.schema import (
    CreateTableStatement,
    GraphsenseSchemas,
    create_parser,
    remove_eol_comments,
    types,
)


def test_create_parser_pkrow(capsys):
    create_parser.parse(
        "create table if not exists test.bla   (a b , b c, primary key (a)) with test;"
    )


def test_create_parser_nokey(capsys):
    ret = create_parser.parse("create table if not exists test.bla   (a b, c d)")

    assert ret["table"]["table"] == "bla"
    assert ret["table"]["keyspace"] == "test"
    assert ret["ine"] is True


def test_create_parser_nokeyNoIne(capsys):
    ret = create_parser.parse("create table test_demo (a b, c d)")

    assert ret["table"]["table"] == "test_demo"
    assert ret["ine"] is not True
    assert ret["withstmt"] is None


def test_create_parser_WithPK(capsys):
    ret = create_parser.parse(
        "create table test_demo (a_b b, c d primary key) with a=b and c={a:bn}"
    )

    assert ret["table"]["table"] == "test_demo"
    assert ret["ine"] is not True
    assert ret["withstmt"] == "a=b and c={a:bn}"
    assert ret["columns"][0]["pk"] is not True
    assert ret["columns"][0]["name"] == "a_b"
    assert ret["columns"][0]["type"] == "b"
    assert ret["columns"][1]["pk"] is True
    assert ret["columns"][1]["name"] == "c"
    assert ret["columns"][1]["type"] == "d"


def test_create_parser_PkSpaces(capsys):
    ret = create_parser.parse(
        "create table test_demo (a b, c d primary key) with a=b and c={a:bn}"
    )

    assert ret["table"]["table"] == "test_demo"
    assert ret["ine"] is not True
    assert ret["withstmt"] == "a=b and c={a:bn}"


def test_create_parser_compoundkey(capsys):
    ret = create_parser.parse(
        "create table test_demo (a b, c d, primary key (a, b)) with a=b and c={a:bn}"
    )

    assert ret["table"]["table"] == "test_demo"
    assert ret["ine"] is not True
    assert ret["withstmt"] == "a=b and c={a:bn}"

    assert ret["columns"][2]["pk"] is True
    assert ret["columns"][2]["compound_key"] == ["a", "b"]


def test_create_parser_compoundkeycomplex(capsys):
    ret = create_parser.parse(
        "CREATE table test_demo "
        "(a b, c d, primary key ((a, b,c), b)) with a=b and c={a:bn}"
    )

    assert ret["table"]["table"] == "test_demo"
    assert ret["ine"] is not True
    assert ret["withstmt"] == "a=b and c={a:bn}"

    assert ret["columns"][2]["pk"] is True
    assert ret["columns"][2]["compound_key"] == [["a", "b", "c"], "b"]


def test_CreateTableStatement_pkrow(capsys):
    CreateTableStatement.from_schema(
        "CREATE TABLE if not exists test.bla   (a b , b c, primary key (a)) with test;"
    )


def test_CreateTableStatement_nokey(capsys):
    ret = CreateTableStatement.from_schema(
        "create table if not exists test.bla   (a b, c d)"
    )

    assert ret.pk == []
    assert ret.columns == {"a": "b", "c": "d"}
    assert ret.if_not_exists is True
    assert ret.keyspace == "test"
    assert ret.table == "bla"
    assert ret.with_stmt is None


def test_CreateTableStatement_nokeyNoIne(capsys):
    ret = CreateTableStatement.from_schema("create table test_demo (a b, c d)")

    assert ret.pk == []
    assert ret.columns == {"a": "b", "c": "d"}
    assert ret.if_not_exists is False
    assert ret.keyspace is None
    assert ret.table == "test_demo"
    assert ret.with_stmt is None


def test_CreateTableStatement_WithPK(capsys):
    ret = CreateTableStatement.from_schema(
        "create table if    not   exists  blub.test_demo "
        "(a_b b, c d primary key) with a=b and c={a:bn}"
    )

    assert ret.pk == ["c"]
    assert ret.columns == {"a_b": "b", "c": "d"}
    assert ret.if_not_exists is True
    assert ret.keyspace == "blub"
    assert ret.table == "test_demo"
    assert ret.with_stmt == "a=b and c={a:bn}"


def test_CreateTableStatement_compoundkey(capsys):
    ret = CreateTableStatement.from_schema(
        "create table test_demo (a b, c d, primary key (a, b)) with a=b and c={a:bn}"
    )

    assert ret.pk == ["a", "b"]
    assert ret.columns == {"a": "b", "c": "d"}
    assert ret.if_not_exists is False
    assert ret.keyspace is None
    assert ret.table == "test_demo"
    assert ret.with_stmt == "a=b and c={a:bn}"


def test_CreateTableStatement_compoundkeycomplex(capsys):
    ret = CreateTableStatement.from_schema(
        "create table test_demo "
        "(a b, c d, primary key ((a, b,c), b)) with a=b and c={a:bn};   "
    )

    assert ret.pk == [["a", "b", "c"], "b"]
    assert ret.columns == {"a": "b", "c": "d"}
    assert ret.if_not_exists is False
    assert ret.keyspace is None
    assert ret.table == "test_demo"
    assert ret.with_stmt == "a=b and c={a:bn}"


def test_CreateTableStatement_strrepr(capsys):
    ret = CreateTableStatement.from_schema(
        "create table test_demo (a b, c d, primary key ((a, b,c), b))"
    )

    assert repr(ret) == "CREATE TABLE test_demo (a b, c d, PRIMARY KEY ((a,b,c),b));"


def test_load_and_parse_embedded_schemaFiles():
    schemas = GraphsenseSchemas().get_all_schemas()

    for f, s in schemas:
        stmts = s.parse_create_table_statements()

        for rs in stmts:
            rstmt = CreateTableStatement.from_schema(repr(rs))
            assert rs == rstmt


def test_get_schema_by_type():
    schemas = GraphsenseSchemas().get_by_schema_type("account_trx", "transformed")
    assert len(schemas) > 1

    schemas = GraphsenseSchemas().get_by_schema_type("account", "transformed")
    assert len(schemas) > 1

    schemas = GraphsenseSchemas().get_by_currency(
        "eth", "transformed", no_extensions=True
    )
    assert len(schemas) == 1

    schemas = GraphsenseSchemas().get_by_currency(
        "trx", "transformed", no_extensions=True
    )
    assert len(schemas) == 1


def test_types_parsing_map():
    res = types.parse("map<integer, float>")
    assert res == "map<integer, float>"


def test_types_parsing_listfrozenoutput():
    res = types.parse("list<FROZEN<tx_input_output>>")
    assert res == "list<frozen<tx_input_output>>"


def test_remove_inline_comments():
    schema = """
        CREATE TABLE trc10 (
            owner_address blob,
            name text,
            abbr text,
            total_supply varint, -- bigint probably barely enough, varint just to be safe
            trx_num varint, -- int probably barely enough, varint just to be safe
            num varint, -- int probably barely enough, varint just to be safe
            start_time varint, -- removed 3 last digits to conform with eth, not always zeros, but can be large 3 outliers
            end_time varint, -- removed 3 last digits to conform with eth, not always zeros, multiple outliers
            description text,
            url text,
            id int,
            frozen_supply list<FROZEN<trc10_frozen_supply>>,
            public_latest_free_net_time varint,
            vote_score smallint,
            free_asset_net_limit bigint,
            public_free_asset_net_limit bigint,
            precision smallint,
            PRIMARY KEY (id)
        );
    """

    schema_target = """
        CREATE TABLE trc10 (
            owner_address blob,
            name text,
            abbr text,
            total_supply varint,
            trx_num varint,
            num varint,
            start_time varint,
            end_time varint,
            description text,
            url text,
            id int,
            frozen_supply list<FROZEN<trc10_frozen_supply>>,
            public_latest_free_net_time varint,
            vote_score smallint,
            free_asset_net_limit bigint,
            public_free_asset_net_limit bigint,
            precision smallint,
            PRIMARY KEY (id)
        );
    """

    assert remove_eol_comments(schema) == schema_target


def test_schema_trx_loads():
    schema = [
        s
        for (n, s) in GraphsenseSchemas().get_by_schema_type(
            "account_trx", "transformed"
        )
        if n == "transformed_account_schema.sql"
    ]
    assert len(schema) == 1
    schema = schema[0]

    create_t_bt = [
        x
        for x in schema.parse_create_table_statements()
        if x.table == "block_transactions"
    ]
    assert len(create_t_bt) == 1
    assert create_t_bt[0].columns["tx_id"] == "bigint"



import logging

from .abstract import EventNotifier, FlowEvent, WatchConfig

logger = logging.getLogger(__name__)


class LoggingEventNotifier(EventNotifier):
    def __init__(self):
        self.msgs = []

    def add_notification(self, flow: FlowEvent, receiver_config: WatchConfig, raw_tx):
        self.msgs.append(f"{receiver_config.email}, {flow}, {raw_tx}")

    def send_notifications(self):
        for msg in self.msgs:
            logger.info(self.msgs)
        self.msgs = []



"""
Transaction Class. Functionality depending on the address type.
"""

from typing import Union

from ..utils import hex_to_bytes


class TransactionHashUtxo:
    def __init__(self, txhash: Union[str, bytearray], config):
        raise NotImplementedError("TransactionUtxo not implemented yet")


class TransactionHashAccount:
    def __init__(self, txhash: Union[str, bytearray], config):
        """Init a transaction instance.

        Args:
            txhash (Union[str, bytearray]): transaction hash
            config (ConfigRow): entry from the config table in the transformed keyspace

        Raises:
            Exception: Description
            ValueError: Description
        """
        self.prefix_length = int(config.tx_prefix_length)
        if type(txhash) == str:
            self.tx_hash_bytes = hex_to_bytes(txhash)
        elif type(txhash) == bytearray:
            self.tx_hash_bytes = txhash
        elif type(txhash) == bytes:
            self.tx_hash_bytes = bytearray(txhash)
        else:
            raise Exception(f"Unknown type for txhash type: {type(txhash)}")

        # todo potentially different length for tron, also saved in bytearray though
        if len(self.tx_hash_bytes) != 32:
            raise ValueError(
                f"Address is not the right length {len(self.tx_hash_bytes)}"
            )

    @property
    def hex(self) -> str:  # noqa
        return self.tx_hash_bytes.hex()

    @property
    def db_encoding(self) -> str:
        return self.bytearray
        # return f"0x{self.hex}"

    @property
    def db_encoding_query(self) -> str:
        return f"0x{self.hex}"

    @property
    def prefix(self) -> str:
        return self.hex.upper()[: self.prefix_length]

    @property
    def bytearray(self) -> bytearray:  # noqa
        return self.tx_hash_bytes



import logging
from datetime import datetime
from typing import List

import pandas as pd
import requests
from simplejson.errors import JSONDecodeError

from ..db import DbFactory
from ..db.analytics import DATE_FORMAT

MIN_START = "2010-10-17"  # no CoinDesk exchange rates available before
logger = logging.getLogger(__name__)


def fetch_exchange_rates(
    start_date: str, end_date: str, symbol_list: List
) -> pd.DataFrame:
    """Fetch BTC exchange rates from CoinDesk.

    Parameters
    ----------
    start_date : str
        Start date (ISO-format YYYY-mm-dd).
    end_date : str
        End date (ISO-format YYYY-mm-dd).
    symbol_list: list[str]
        ["EUR", "USD", "JPY" ...]

    Returns
    -------
    DataFrame
        Exchange rates in pandas DataFrame with columns 'date', 'fiat_values'
    """
    df_merged = pd.DataFrame()

    for fiat in symbol_list:
        url = (
            f"https://api.coindesk.com/v1/bpi/historical/close.json"
            f"?currency={fiat.lower()}&start={start_date}&end={end_date}"
        )
        logger.info(f"Fetching url: {url}")
        request = requests.get(url)
        try:
            json = request.json()
            logger.info(json["disclaimer"])

            if "bpi" not in json:
                # API of coindesk does not deliver recent rates anymore
                # current last rates avail where till 07-07-2022 (on 14-10-2020)
                logger.warning(
                    "No exchange rates found for "
                    f"{fiat} in range {start_date} - {end_date}"
                )
                continue

            rates = pd.DataFrame.from_records([json["bpi"]]).transpose()
            rates.rename(columns={0: fiat}, inplace=True)
            df_merged = rates.join(df_merged)
        except JSONDecodeError as symbol_not_found:
            logger.error(f"Unknown currency: {fiat}")
            raise SystemExit(1) from symbol_not_found

    df_merged.reset_index(level=0, inplace=True)
    df_merged.rename(columns={"index": "date"}, inplace=True)
    df_merged["fiat_values"] = df_merged.drop("date", axis=1).to_dict(orient="records")

    [
        df_merged.drop(c, axis=1, inplace=True)
        for c in symbol_list
        if c in df_merged.keys()
    ]
    return df_merged


def fetch(env, currency, fiat_currencies, start_date, end_date):
    with DbFactory().from_config(env, currency) as db:
        return fetch_impl(
            db,
            env,
            currency,
            fiat_currencies,
            start_date,
            end_date,
            None,
            start_date != MIN_START,
            False,
            False,
        )


def fetch_impl(
    db,
    env,
    currency,
    fiat_currencies,
    start_date,
    end_date,
    table,
    force,
    dry_run,
    abort_on_gaps,
):
    if datetime.fromisoformat(start_date) < datetime.fromisoformat(MIN_START):
        logger.warning(f"Warning: Exchange rates not available before {MIN_START}")
        start_date = MIN_START

    # query most recent data in 'exchange_rates' table
    if not force:
        logger.info(f"Get last imported rate from {db.raw.get_keyspace()}")
        most_recent_date = db.raw.get_last_exchange_rate_date(table=table)
        if most_recent_date is not None:
            start_date = most_recent_date.strftime(DATE_FORMAT)

    logger.info(f"*** Starting exchange rate ingest for {currency} ***")
    logger.info(f"Start date: {start_date}")
    logger.info(f"End date: {end_date}")

    if datetime.fromisoformat(start_date) > datetime.fromisoformat(end_date):
        logger.error("Error: start date after end date.")
        raise SystemExit(1)

    logger.info(f"Target fiat currencies: {fiat_currencies}")
    exchange_rates = fetch_exchange_rates(start_date, end_date, fiat_currencies)
    return exchange_rates


def ingest(
    env,
    currency,
    fiat_currencies,
    start_date,
    end_date,
    table,
    force,
    dry_run,
    abort_on_gaps,
):
    if dry_run:
        logger.warning("This is a Dry-Run. Nothing will be written to the database!")
    with DbFactory().from_config(env, currency) as db:
        exchange_rates = fetch_impl(
            db,
            env,
            currency,
            fiat_currencies,
            start_date,
            end_date,
            table,
            force,
            dry_run,
            abort_on_gaps,
        )

        # insert exchange rates into Cassandra table
        if not dry_run:
            logger.info(f"Writing to keyspace {db.raw.get_keyspace()}")
            if len(exchange_rates) > 0:
                db.raw.ingest(table, exchange_rates.to_dict("records"))
                logger.info(f"Inserted rates for {len(exchange_rates)} days: ")
                logger.info(
                    f"{exchange_rates.iloc[0].date} - {exchange_rates.iloc[-1].date}"
                )
            else:
                logger.info("Nothing to insert.")
        else:
            logger.info(
                "Dry run: No data inserted. "
                f"Would have inserted {len(exchange_rates)} days."
            )
            logger.info(exchange_rates)



from cashaddress.convert import Address, to_legacy_address  # InvalidAddress

# Patch P2SH32 legacy address support
# https://bch.info/en/upgrade
if ("P2SH32", 5, False) not in Address.VERSION_MAP["legacy"]:
    Address.VERSION_MAP["legacy"].append(("P2SH32", 5, False))
if ("P2SH32", 11, False) not in Address.VERSION_MAP["cash"]:
    Address.VERSION_MAP["cash"].append(("P2SH32", 11, False))


def bch_address_to_legacy(address):
    return to_legacy_address(address)



import logging
import time
from typing import Dict, List, Optional, Tuple

import pandas as pd
from diskcache import Cache

from graphsenselib.config.config import DeltaUpdaterConfig
from graphsenselib.db import DbChange
from graphsenselib.deltaupdate.update.abstractupdater import (
    TABLE_NAME_DELTA_HISTORY,
    UpdateStrategy,
)
from graphsenselib.deltaupdate.update.account.createchanges import (
    get_bookkeeping_changes,
    prepare_balances_for_ingest,
    prepare_entities_for_ingest,
    prepare_entity_txs_for_ingest,
    prepare_relations_for_ingest,
    prepare_txs_for_ingest,
)
from graphsenselib.deltaupdate.update.account.createdeltas import (
    get_balance_deltas,
    get_entity_transaction_updates_trace_token,
    get_entity_transactions_updates_tx,
    get_entity_updates_trace_token,
    get_entity_updates_tx,
    get_sorted_unique_addresses,
    only_call_traces,
    relationdelta_from_tokentransfer,
    relationdelta_from_trace,
    relationdelta_from_transaction,
)
from graphsenselib.deltaupdate.update.account.modelsdelta import (
    BalanceDelta,
    DbDeltaAccount,
    RawEntityTxAccount,
    RelationDeltaAccount,
)
from graphsenselib.deltaupdate.update.account.modelsraw import (
    AccountBlockAdapter,
    AccountLogAdapter,
    AccountTransactionAdapter,
    Block,
    EthTraceAdapter,
    Log,
    Trace,
    Transaction,
    TrxTraceAdapter,
    TrxTransactionAdapter,
)
from graphsenselib.deltaupdate.update.account.tokens import ERC20Decoder
from graphsenselib.deltaupdate.update.generic import Action, ApplicationStrategy, Tx
from graphsenselib.deltaupdate.update.utxo.update import apply_changes
from graphsenselib.schema.schema import GraphsenseSchemas
from graphsenselib.utils import DataObject as MutableNamedTuple
from graphsenselib.utils import no_nones
from graphsenselib.utils.account import (
    get_id_group_with_secondary_addresstransactions,
    get_id_group_with_secondary_relations,
)
from graphsenselib.utils.cache import TableBasedCache
from graphsenselib.utils.errorhandling import CrashRecoverer
from graphsenselib.utils.logging import LoggerScope

logger = logging.getLogger(__name__)
inout_logger = logging.getLogger("inout_logger")


COINBASE_PSEUDO_ADDRESS = None
PSEUDO_ADDRESS_AND_IDS = {COINBASE_PSEUDO_ADDRESS: -1}

DEFAULT_SUMMARY_STATISTICS = MutableNamedTuple(
    **{
        "id": 0,
        "timestamp": 0,
        "timestamp_transform": 0,
        "no_blocks": 0,
        "no_blocks_transform": 0,
        "no_transactions": 0,
        "no_addresses": 0,
        "no_address_relations": 0,
        "no_clusters": 0,
        "no_cluster_relations": 0,
    }
)


class UpdateStrategyAccount(UpdateStrategy):
    def __init__(
        self,
        db,
        du_config: DeltaUpdaterConfig,
        pedantic: bool,
        application_strategy: ApplicationStrategy = ApplicationStrategy.TX,
        patch_mode: bool = False,
        forward_fill_rates: bool = False,
    ):
        super().__init__(db, du_config.currency, forward_fill_rates=forward_fill_rates)
        self.du_config = du_config
        crash_file = (
            "/tmp/account_deltaupdate_"
            f"{self._db.raw.get_keyspace()}_{self._db.transformed.get_keyspace()}"
            "_crashreport.err"
        )
        stats_value = self._db.transformed.get_summary_statistics()
        """ Make statistics row mutable"""
        self._statistics = (
            MutableNamedTuple(**stats_value._asdict())
            if stats_value is not None
            else DEFAULT_SUMMARY_STATISTICS
        )
        # get ingest config
        self._pedantic = pedantic
        self._patch_mode = patch_mode
        self.changes = None
        self.application_strategy = application_strategy
        logger.info(f"Updater running in {application_strategy} mode.")
        self.crash_recoverer = CrashRecoverer(crash_file)

    def consume_transaction_id_composite(self, block_id, transaction_index):
        return (block_id << 32) + transaction_index

    def clear_cache(self):
        cache = Cache(self.du_config.fs_cache.directory)
        cache.clear()

    def clear_cache_for_blocks(self, blocks: List[int]):
        cache = TableBasedCache(Cache(self.du_config.fs_cache.directory))
        for block in blocks:
            cache.delete_item("block", block)
            cache.delete_item("trace", block)
            cache.delete_item("log", block)

            if self.currency == "trx":
                transactions = cache.get(("transaction", block), [])
                for tx in transactions:
                    cache.delete_item("fee", tx["tx_hash"])

            cache.delete_item("transaction", block)

    def persist_updater_progress(self):
        if self.changes is not None:
            atomic = ApplicationStrategy.TX == self.application_strategy
            apply_changes(
                self._db, self.changes, self._pedantic, try_atomic_writes=atomic
            )
            self.changes = None
        self._time_last_batch = time.time() - self._batch_start_time

    def prepare_database(self):
        with LoggerScope.debug(logger, "Preparing database"):
            if self._db.transformed.has_delta_updater_v1_tables():
                raise Exception(
                    "Tables of the delta-updater v1 detected. "
                    "please delete new_addresses, dirty_address, "
                    "detla_updater_state and delta-updater_history "
                    "before using delta updater v2."
                )
            GraphsenseSchemas().ensure_table_exists_by_name(
                self._db.transformed,
                TABLE_NAME_DELTA_HISTORY,
                truncate=False,
            )

    def get_block_data(self, cache, block):
        blocks = cache.get(("block", block), [])
        txs = cache.get(("transaction", block), [])
        traces = cache.get(("trace", block), [])
        logs = cache.get(("log", block), [])
        return txs, traces, logs, blocks

    def get_fee_data(self, cache, tx):
        return cache.get(("fee", tx), [{"fee": None}])[0]["fee"]

    def process_batch_impl_hook(self, batch: List[int]) -> Tuple[Action, Optional[int]]:
        rates = {}
        transactions = []
        traces = []
        logs = []
        blocks = []
        bts = {}
        """
            Read transaction and exchange rates data
        """

        with LoggerScope.debug(logger, "Checking recovery state.") as lg:
            if self.crash_recoverer.is_in_recovery_mode():
                """
                If we are in recovery mode we start with a block earlier to catch up
                the delta otherwise would start with whats in the db +1
                In case of an error in between blocks this would mean skipping to
                the next block
                """
                mb = max(0, min(batch) - 1)
                lg.warning(
                    "Delta update is in crash recovery mode. Crash hint is "
                    f"{self.crash_recoverer.get_recovery_hint()} in "
                    f"{self.crash_recoverer.get_recovery_hint_filename()} "
                    f" restarting at block {mb}."
                )
                batch = [mb] + batch

        with LoggerScope.debug(logger, "Reading transaction and rates data") as log:
            missing_rates_in_block = False
            cache = TableBasedCache(Cache(self.du_config.fs_cache.directory))
            for block in batch:
                txs_new, traces_new, logs_new, blocks_new = self.get_block_data(
                    cache, block
                )

                if len(blocks_new) == 0:  #
                    msg = (
                        f"Block {block} is not present in cache. Please ingest"
                        f" more blocks with option --sinks fs-cache"
                    )
                    log.error(msg)

                    if block == batch[0]:
                        log.warning("First block of batch not in cache.")
                        return Action.BREAK, None
                    else:
                        log.warning(
                            "Skipping block without data in cache."
                            "Continuing with next block."
                        )
                        break

                final_block = block

                transactions.extend(txs_new)
                traces.extend(traces_new)
                logs.extend(logs_new)
                blocks.extend(blocks_new)

                fiat_values = self._db.transformed.get_exchange_rates_by_block(
                    block
                ).fiat_values
                if fiat_values is None:
                    missing_rates_in_block = True
                    fiat_values = [0, 0]
                rates[block] = fiat_values
                bts[block] = self._db.raw.get_block_timestamp(block)

            if missing_rates_in_block:
                log.warning("Block Range has missing exchange rates. Using Zero.")

        if self.application_strategy == ApplicationStrategy.BATCH:
            if self.crash_recoverer.is_in_recovery_mode():
                raise Exception("Batch mode is not allowed in recovery mode.")

            if self.currency == "trx":
                trace_adapter = TrxTraceAdapter()
                transaction_adapter = TrxTransactionAdapter()
                for tx in transactions:
                    tx["fee"] = self.get_fee_data(cache, tx["tx_hash"])

            elif self.currency == "eth":
                trace_adapter = EthTraceAdapter()
                transaction_adapter = AccountTransactionAdapter()

            # convert dictionaries to dataclasses and unify naming
            log_adapter = AccountLogAdapter()
            block_adapter = AccountBlockAdapter()
            traces = trace_adapter.dicts_to_renamed_dataclasses(traces)
            traces = trace_adapter.process_fields_in_list(traces)
            transactions = transaction_adapter.dicts_to_dataclasses(transactions)
            logs = log_adapter.dicts_to_dataclasses(logs)
            blocks = block_adapter.dicts_to_dataclasses(blocks)
            blocks = block_adapter.process_fields_in_list(blocks)

            changes = []

            (tx_changes, nr_new_addresses, nr_new_address_relations) = self.get_changes(
                transactions, traces, logs, blocks, rates
            )

            changes.extend(tx_changes)

            if self.currency == "trx":
                nr_new_tx = len([tx for tx in transactions if tx.receipt_status == 1])
            else:
                nr_new_tx = len(transactions)

            runtime_seconds = int(time.time() - self.batch_start_time)

            bookkeeping_changes = get_bookkeeping_changes(
                self._statistics,
                self._db.transformed.get_summary_statistics(),
                final_block,
                nr_new_address_relations,
                nr_new_addresses,
                nr_new_tx,
                self.highest_address_id,
                runtime_seconds,
                bts,
                len(blocks),
                patch_mode=self._patch_mode,
            )

            changes.extend(bookkeeping_changes)

            # Store changes to be written
            # They are applied at the end of the batch in
            # persist_updater_progress
            self.changes = changes
            return Action.CONTINUE, final_block

        else:
            raise ValueError(
                f"Unknown application strategy {self.application_strategy}"
            )

    def get_changes(
        self,
        transactions: List[Transaction],
        traces: List[Trace],
        logs: List[Log],
        blocks: List[Block],
        rates: Dict[int, List],
    ) -> Tuple[List[DbChange], int, int]:
        currency = self.currency.upper()
        id_bucket_size = self._db.transformed.get_address_id_bucket_size()
        block_bucket_size = self._db.transformed.get_block_id_bucket_size()
        block_bucket_size_address_txs = (
            self._db.transformed.get_address_transactions_id_bucket_size()
        )
        relations_nbuckets = self._db.transformed.get_addressrelations_ids_nbuckets()
        tdb = self._db.transformed

        def get_next_address_ids_with_aliases(address: str):
            return (
                self.consume_address_id()
                if address not in PSEUDO_ADDRESS_AND_IDS
                else PSEUDO_ADDRESS_AND_IDS[address]
            )

        def get_tx_prefix(tx_hash):
            tx_hash = tdb.to_db_tx_hash(tx_hash)
            return (tx_hash.db_encoding, tx_hash.prefix)

        def get_address_prefix(address_str):
            address = tdb.to_db_address(address_str)
            return (address.db_encoding, address.prefix)

        hash_to_id = {
            tx.tx_hash: self.consume_transaction_id_composite(
                tx.block_id, tx.transaction_index
            )
            for tx in transactions
        }

        if currency == "TRX":
            transactions = [tx for tx in transactions if tx.to_address is not None]
            transactions = [tx for tx in transactions if tx.receipt_status == 1]

        elif currency == "ETH":
            pass
        else:
            raise ValueError(f"Unknown currency {currency}")

        tx_hashes = [tx.tx_hash for tx in transactions]
        reward_traces = [t for t in traces if t.tx_hash is None]
        # traces without reward traces:
        traces = [trace for trace in traces if trace.tx_hash is not None]
        # calculate successful traces:
        traces_s = [trace for trace in traces if trace.status == 1]

        hash_to_tx = dict(zip(tx_hashes, transactions))

        with LoggerScope.debug(logger, "Decode logs to token transfers"):
            supported_tokens = self._db.transformed.get_token_configuration()
            tokendecoder = ERC20Decoder(currency, supported_tokens)
            token_transfers = no_nones(
                [tokendecoder.log_to_transfer(log) for log in logs]
            )

        with LoggerScope.debug(logger, "Compute unique addresses in correct order"):
            if currency == "TRX":
                transactions_for_addresses = transactions
            elif currency == "ETH":
                transactions_for_addresses = []
            else:
                raise ValueError(f"Unknown currency {currency}")

            addresses = get_sorted_unique_addresses(
                traces_s, reward_traces, token_transfers, transactions_for_addresses
            )
            len_addr = len(addresses)

        with LoggerScope.debug(
            logger, f"Checking existence for {len_addr} addresses"
        ) as _:
            addr_ids = dict(tdb.get_address_id_async_batch(list(addresses)))

        with LoggerScope.debug(logger, "Reading addresses to be updated"):
            existing_addr_ids = no_nones(
                [address_id.result_or_exc.one() for adr, address_id in addr_ids.items()]
            )

            global addresses_resolved
            addresses_resolved = dict(
                tdb.get_address_async_batch(
                    [adr.address_id for adr in existing_addr_ids]
                )
            )

            def get_resolved_address(addr_id_exc):
                addr_id = addr_id_exc.result_or_exc.one()
                return (
                    (None, None)
                    if addr_id is None
                    else (
                        addr_id.address_id,
                        addresses_resolved[
                            addr_id.address_id
                        ].result_or_exc.one(),  # noqa
                    )
                )

            addresses_to_id__rows = {
                adr: get_resolved_address(address_id)
                for adr, address_id in addr_ids.items()
            }

            del addresses_resolved

            bytes_to_row_address = {
                address: row[1] for address, row in addresses_to_id__rows.items()
            }
            for addr in addresses:
                addr_id, address = addresses_to_id__rows[addr]
                if addr_id is None:
                    new_addr_id = get_next_address_ids_with_aliases(addr)
                    addresses_to_id__rows[addr] = (new_addr_id, None)

            address_hash_to_id = {
                address: id_row[0] for address, id_row in addresses_to_id__rows.items()
            }

        with LoggerScope.debug(logger, "Get transactions to insert into the database"):
            txs_to_insert = []

            for tx_hash in tx_hashes:
                tx_id = hash_to_id[tx_hash]
                tx_index = hash_to_tx[tx_hash].transaction_index
                block_id = hash_to_tx[tx_hash].block_id
                failed = hash_to_tx[tx_hash].receipt_status == 0
                txs_to_insert.append(
                    Tx(
                        block_id=block_id,
                        tx_id=tx_id,
                        tx_hash=tx_hash,
                        tx_index=tx_index,
                        failed=failed,
                    )
                )

        with LoggerScope.debug(
            logger, "Get entity transaction updates - traces and tokens"
        ):
            entity_transactions = []
            entity_deltas = []

            if currency == "TRX":
                traces_s_filtered = only_call_traces(traces_s)  # successful and call
            elif currency == "ETH":
                traces_s_filtered = traces_s
            else:
                raise ValueError(f"Unknown currency {currency}")

            entity_transactions += get_entity_transaction_updates_trace_token(
                traces_s_filtered,
                token_transfers,
                hash_to_id,
                address_hash_to_id,
                rates,
            )

        with LoggerScope.debug(logger, "Get entity updates - traces and tokens"):
            entity_deltas += get_entity_updates_trace_token(
                traces_s_filtered,
                token_transfers,
                reward_traces,
                hash_to_id,
                currency,
                rates,
            )

        with LoggerScope.debug(logger, "Get relation updates - traces and tokens"):
            relation_updates_trace = [
                relationdelta_from_trace(trace, rates, currency)
                for trace in traces_s_filtered
            ]
            relation_updates_tokens = [
                relationdelta_from_tokentransfer(tt, rates) for tt in token_transfers
            ]
            relation_updates = relation_updates_trace + relation_updates_tokens

        with LoggerScope.debug(
            logger,
            "Get entity and entity transaction "
            "updates from tranasactions (only tron)",
        ):
            # in eth we disregard the eth values because they are already in the traces
            # in tron only traces that are not the initial transaction have values,
            # so we still need to add the value from the transaction
            if currency == "TRX":
                entity_transactions_tx = get_entity_transactions_updates_tx(
                    transactions, hash_to_id, address_hash_to_id
                )

                entity_deltas_tx = get_entity_updates_tx(
                    transactions,
                    hash_to_id,
                    currency,
                    rates,
                )

                entity_deltas += entity_deltas_tx

                relation_updates_tx = [
                    relationdelta_from_transaction(tx, rates, currency)
                    for tx in transactions
                    if tx.from_address is not None
                ]

                entity_transactions += entity_transactions_tx
                relation_updates += relation_updates_tx

            elif currency == "ETH":
                relation_updates_tx = []
            else:
                raise ValueError(f"Unknown currency {currency}")

        with LoggerScope.debug(logger, "Get balance updates"):
            """Get balance updates"""
            balance_updates = get_balance_deltas(
                relation_updates_trace,
                relation_updates_tx,
                relation_updates_tokens,
                reward_traces,
                transactions,
                blocks,
                address_hash_to_id,
                currency,
            )

        with LoggerScope.debug(logger, "Create dbdelta and compress"):
            """Combine all updates except the pure inserts into a delta object"""
            dbdelta = DbDeltaAccount(
                entity_deltas, entity_transactions, relation_updates, balance_updates
            )
            """ Group and merge deltas before merge with db deltas """
            dbdelta = dbdelta.compress()

        with LoggerScope.debug(logger, "Query data from database"):
            # Query outrelations
            rel_to_query = [
                (
                    addresses_to_id__rows[update.src_identifier][0],
                    addresses_to_id__rows[update.dst_identifier][0],
                )
                for update in dbdelta.relation_updates
            ]
            addr_outrelations_q = tdb.get_address_outrelations_async_batch_account(
                rel_to_query
            )
            addr_outrelations = {
                (update.src_identifier, update.dst_identifier): qr
                for update, qr in zip(dbdelta.relation_updates, addr_outrelations_q)
            }

            # Query inrelations
            rel_to_query = [
                (
                    addresses_to_id__rows[update.dst_identifier][0],
                    addresses_to_id__rows[update.src_identifier][0],
                )
                for update in dbdelta.relation_updates
            ]
            addr_inrelations_q = tdb.get_address_inrelations_async_batch_account(
                rel_to_query
            )
            addr_inrelations = {
                (update.src_identifier, update.dst_identifier): qr
                for update, qr in zip(dbdelta.relation_updates, addr_inrelations_q)
            }

            # Query balances of addresses
            address_ids = [addresses_to_id__rows[address][0] for address in addresses]
            addr_balances_q = (
                tdb.get_balance_async_batch_account(  # could probably query less
                    address_ids
                )
            )
            addr_balances = {
                addr_id: BalanceDelta.from_db(addr_id, qr.result_or_exc.all())
                for addr_id, qr in zip(address_ids, addr_balances_q)
            }

        with LoggerScope.debug(logger, "Prepare changes"):
            changes = []

            """ Inserts of new transactions """
            changes += prepare_txs_for_ingest(
                txs_to_insert,
                id_bucket_size,
                block_bucket_size,
                get_tx_prefix,
            )

            """ Merging max secondary ID """
            changes += self.prepare_and_query_max_secondary_id(
                dbdelta.relation_updates,
                dbdelta.new_entity_txs,
                id_bucket_size,
                address_hash_to_id,
            )

            """ Merging entity transactions """
            changes += prepare_entity_txs_for_ingest(
                dbdelta.new_entity_txs,
                id_bucket_size,
                currency,
                block_bucket_size_address_txs,
            )

            """ Merging balances"""
            changes += prepare_balances_for_ingest(
                dbdelta.balance_updates, id_bucket_size, addr_balances
            )

            """ Merging relations deltas """
            (
                changes_relations,
                new_rels_in,
                new_rels_out,
            ) = prepare_relations_for_ingest(
                dbdelta.relation_updates,
                address_hash_to_id,
                addr_inrelations,
                addr_outrelations,
                id_bucket_size,
                relations_nbuckets,
            )
            changes += changes_relations

            """ Merging entity deltas """
            entity_changes, nr_new_entities = prepare_entities_for_ingest(
                dbdelta.entity_updates,
                address_hash_to_id,
                bytes_to_row_address,
                new_rels_in,
                new_rels_out,
                id_bucket_size,
                get_address_prefix,
            )
            changes += entity_changes

            assert sum(new_rels_in.values()) == sum(new_rels_out.values())
            nr_new_rels = sum(new_rels_in.values())
            nr_new_entities_created = nr_new_entities

        return (
            changes,
            nr_new_entities_created,
            nr_new_rels,
        )

    def prepare_and_query_max_secondary_id(
        self,
        relation_updates: List[RelationDeltaAccount],
        new_entity_txs: List[RawEntityTxAccount],
        id_bucket_size: int,
        address_hash_to_id: Dict[bytes, int],
    ):
        relations_nbuckets = self._db.transformed.get_addressrelations_ids_nbuckets()

        def max_secondary_dict_from_db(df, id_group_col, grp_col):
            # query max secondary ids from database
            unique_address_id_groups = list(df[grp_col])
            max_secondary_atx = self._db.transformed.get_max_secondary_ids_async(
                unique_address_id_groups, tablename, id_group_col
            )
            max_secondary_atx = [qr.result_or_exc.one() for qr in max_secondary_atx]
            # use placeholder -1 if there is nothing in the database yet.
            # Will be consumed by max
            # and not 0, or otherwise the logic will say it shouldnt update
            max_secondary_atx = [
                res[1] if res is not None else -1 for res in max_secondary_atx
            ]
            max_secondary_atx = dict(zip(unique_address_id_groups, max_secondary_atx))
            return max_secondary_atx

        def get_max_secondary_changes(data, tablename, grp_col, sec_col):
            max_col = "max_secondary_id"
            df = pd.DataFrame(data, columns=[grp_col, sec_col])
            df = df.groupby(grp_col).max()
            df = df.reset_index()
            df = df.rename(columns={sec_col: max_col})
            max_secondary_atx = max_secondary_dict_from_db(df, grp_col, grp_col)
            df[max_col + "old"] = df[grp_col].map(max_secondary_atx)
            df[max_col] = df[[max_col, max_col + "old"]].max(axis=1)
            # convert to Db changes

            changes = [
                DbChange.update(
                    table=tablename,
                    data={
                        grp_col: row[grp_col],
                        "max_secondary_id": row[max_col],
                    },
                )
                for _, row in df.iterrows()
                if row[max_col] != row[max_col + "old"]
            ]
            return changes

        block_bucket_size_address_txs = (
            self._db.transformed.get_address_transactions_id_bucket_size()
        )
        """ secondary group id for address transactions and address in/out relations"""
        tablename = "address_transactions_secondary_ids"
        grp_col, sec_col = "address_id_group", "address_id_secondary_group"
        secondary_group_data = [
            get_id_group_with_secondary_addresstransactions(
                tx.identifier,
                id_bucket_size,
                tx.block_id,
                block_bucket_size_address_txs,
            )
            for tx in new_entity_txs
        ]
        changes_secondary_atx = get_max_secondary_changes(
            secondary_group_data, tablename, grp_col, sec_col
        )

        tablename = "address_outgoing_relations_secondary_ids"
        grp_col, sec_col = "src_address_id_group", "src_address_id_secondary_group"

        secondary_group_data = [
            get_id_group_with_secondary_relations(
                address_hash_to_id[tx.src_identifier],
                address_hash_to_id[tx.dst_identifier],
                id_bucket_size,
                relations_nbuckets,
            )
            for tx in relation_updates
        ]
        changes_secondary_aor = get_max_secondary_changes(
            secondary_group_data, tablename, grp_col, sec_col
        )

        tablename = "address_incoming_relations_secondary_ids"
        grp_col, sec_col = "dst_address_id_group", "dst_address_id_secondary_group"

        secondary_group_data = [
            get_id_group_with_secondary_relations(
                address_hash_to_id[tx.dst_identifier],
                address_hash_to_id[tx.src_identifier],
                id_bucket_size,
                relations_nbuckets,
            )
            for tx in relation_updates
        ]

        changes_secondary_air = get_max_secondary_changes(
            secondary_group_data, tablename, grp_col, sec_col
        )

        changes = []
        changes += changes_secondary_atx
        changes += changes_secondary_aor
        changes += changes_secondary_air
        return changes



import logging
from datetime import datetime

logger = logging.getLogger(__name__)


def parse_timestamp(timestamp: int) -> datetime:
    check_timestamp(timestamp)
    return datetime.utcfromtimestamp(timestamp)


def check_timestamp(timestamp: int) -> None:
    # check if timestamp is between 2000 and 2100
    if timestamp < 946684800 or timestamp > 4102444800:
        # log warning
        logger.warning(f"Timestamp {timestamp} is not between the years 2000 and 2100")



import multiprocessing as mp
import os
import random
import traceback
from functools import cache
from queue import Empty


class PipelineError:
    def __init__(self, exc, tb, task_name):
        self.exc = exc
        self.tb = tb
        self.task_name = task_name

    def __repr__(self):
        return repr(self.exc) + " @ " + self.task_name


def get_process_graph(pipedef, panic_signal, max_queue_size):
    ips = [mp.Queue(max_queue_size) for _ in pipedef]
    pas = []
    outs = []
    for s, q in zip(pipedef, ips):
        ps, outqs = s.get_flow(q, panic_signal)
        pas.extend(ps)
        outs.extend(outqs)

    return (ips, pas, outs)


def is_stop_signal(msg):
    return isinstance(msg, StopPipeline)


def is_pipline_error(msg):
    return isinstance(msg, PipelineError)


def is_special_signal(msg):
    return is_pipline_error(msg) or is_stop_signal(msg)


def set_context(ctx, *initial_data, **kwargs):
    for dictionary in initial_data:
        for key in dictionary:
            setattr(ctx, key, dictionary[key])
    for key in kwargs:
        setattr(ctx, key, kwargs[key])


class Context:
    pass


class StopPipeline:
    pass


class Pipeline:
    def __init__(
        self, pipedef, collector_task=None, collector_args=(), max_queue_size=0
    ):
        self.pipedef = pipedef
        self.panic_signal = mp.Event()
        ins, ps, outs = get_process_graph(pipedef, self.panic_signal, max_queue_size)
        self.has_joined = False
        if collector_task:
            if isinstance(collector_task, Process):
                if collector_args != ():
                    raise ValueError(
                        "collector_args are not supported for process collector."
                    )
                p, op = collector_task.get_flow(outs, self.panic_signal)
                ps.extend(p)
                outs = op
            else:
                outq = mp.Queue()
                ps.append(
                    mp.Process(
                        target=collector_task,
                        args=(
                            self.panic_signal,
                            outs,
                            outq,
                            *collector_args,
                        ),
                        daemon=True,
                    )
                )
                outs = [outq]

        self.graph = (ins, ps, outs)
        self.stopped = False

    def run(self):
        for p in self._processes:
            p.start()
        return self

    @property
    def is_in_panic_mode(self):
        return self.panic_signal.is_set()

    @property
    def failed(self):
        return self.is_in_panic_mode

    def raise_on_failed(self):
        exs = self.get_errors()
        if len(exs) > 0:
            raise exs[0].exc

    def _join(self):
        if not self.has_joined:
            self.has_joined = True
            for p in self._processes:
                p.join()
        self.stopped = True

    def join(self):
        self.put(StopPipeline())
        self._join()

    def __enter__(self):
        return self.run()

    def get_errors(self):
        res = self.get_results()
        return [ex for ex in res if is_pipline_error(ex)]

    @cache
    def get_results(self):
        if not self.stopped:
            raise ValueError(
                "Get results should only be used after pipeline is stopped."
            )
        res = []
        queues = self.outputs_queues
        while True:
            all_empty = 0
            for q in queues:
                if not q.empty():
                    res.append(q.get())
                else:
                    all_empty += 1
            if all_empty == len(queues):
                break

        return [r for r in res if not is_stop_signal(r)]

    def __exit__(self, exc_type, exc_value, tb):
        self.join()

    @property
    def _inqueues(self):
        ips, _, _ = self.graph
        return ips

    @property
    def _processes(self):
        _, ps, _ = self.graph
        return ps

    def put(self, item):
        if self.is_in_panic_mode:
            item = StopPipeline()
        for q in self._inqueues:
            q.put(item)

    @property
    def outputs_queues(self):
        _, _, ops = self.graph
        return ops


class Process:
    def __init__(
        self, task, init_func=None, init_args=None, teardown_func=None, name="worker"
    ):
        self.init_args = init_args
        self.init_func = init_func
        self.teardown_func = teardown_func
        self.task = task
        self.context = Context()
        self.iname = name
        self.nexts = []

    @property
    def name(self):
        return f"{self.iname} at {os.getpid()}"

    def __or__(self, next_step):
        return self.pipe(next_step)

    def pipe(self, next_step):
        self.nexts.append(next_step)
        return self

    def get_flow(self, inp, panic_signal):
        ps = []
        outsa = []
        if len(self.nexts) > 0:
            outputs = [mp.Queue() for _ in self.nexts]
            for n, q in zip(self.nexts, outputs):
                p, outs = n.get_flow(q, panic_signal)
                ps.extend(p)
                outsa.extend(outs)
        else:
            out = [mp.Queue()]
            return [
                mp.Process(
                    target=self,
                    name=self.name,
                    args=(inp, out, panic_signal),
                    daemon=True,
                )
            ], out

        return (
            ps
            + [
                mp.Process(
                    target=self,
                    name=self.name,
                    args=(inp, outputs, panic_signal),
                    daemon=True,
                )
            ],
            outsa,
        )

    def put_all(self, sinks, data):
        if isinstance(sinks, list):
            for s in sinks:
                s.put(data)
        else:
            sinks.put(data)

    def get_one(self, sources, timeout=1):
        if isinstance(sources, list):
            q = random.choice(sources)
            return q.get(timeout=timeout)
        else:
            return sources.get(timeout=timeout)

    def handle_msg(self, sources, sinks) -> bool:
        """Default handling of messages, get message and process

        Args:
            source (mp.Queue): Description
            sinks (List[mp.Queue]): Description

        Returns:
            bool: if worker should shutdown
        """
        try:
            data = self.get_one(sources, timeout=4)

            if is_special_signal(data):
                self.put_all(sinks, data)
                return True

            if data is not None:
                ret = self.task(self.context, data)
                self.put_all(sinks, ret)
        except Empty:
            return False

        return False

    def handle_msg_panic_mode(self, sources, sinks) -> bool:
        """Panic mode, if some worker had an unhandled exception
        only passes exceptions and stop signals, but does no work anymore.
        Args:
            source (mp.Queue): Description
            sinks (List[mp.Queue]): Description

        Returns:
            bool: if worker should shutdown
        """
        try:
            data = self.get_one(sources)

            if is_special_signal(data):
                self.put_all(sinks, data)
                return True

            return False
        except Empty:
            return False

    def __call__(self, source, sinks, panic_signal):
        if self.init_func:
            self.init_func(self.context, *self.init_args)
        # print(self.name, "ready")
        try:
            while True:
                try:
                    kill_worker = False
                    if not panic_signal.is_set():
                        kill_worker = self.handle_msg(source, sinks)
                    else:
                        kill_worker = self.handle_msg_panic_mode(source, sinks)

                    if kill_worker:
                        break
                except KeyboardInterrupt:
                    pass
                except Exception as e:
                    panic_signal.set()
                    self.put_all(
                        sinks,
                        PipelineError(
                            e, traceback.TracebackException.from_exception(e), self.name
                        ),
                    )
        finally:
            # print(self.name, "i am out")
            if self.teardown_func:
                self.teardown_func(self.context)


class SyncCollector(Process):
    def get_one(self, sources):
        return [q.get() for q in sources]

    def handle_msg(self, sources, sinks) -> bool:
        """Default handling of messages, get message and process

        Args:
            source (mp.Queue): Description
            sinks (List[mp.Queue]): Description

        Returns:
            bool: if worker should shutdown
        """
        try:
            datas = self.get_one(sources)

            ret, end = self.task(self.context, datas)
            if end is None:
                end = False

            if ret is not None:
                self.put_all(sinks, ret)

            return end
        except Empty:
            return False

    def handle_msg_panic_mode(self, sources, sinks) -> bool:
        _ = self.handle_msg(sources, sinks)
        return True



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Tuple

from ..datatypes.common import FlowDirection


@dataclass
class FlowEvent:
    direction: FlowDirection
    address: str
    value: int
    block: int
    timestamp: Optional[datetime]
    tx_ref: Optional[str]


@dataclass
class WatchConfig:
    email: str
    on_incoming: bool
    on_outgoing: bool
    value_gt: Optional[int]


class WatchpointProvider(ABC):
    @abstractmethod
    def is_watched(self, address: str):
        pass

    @abstractmethod
    def get_configuration(self, address: str) -> WatchConfig:
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class EventNotifier(ABC):
    @abstractmethod
    def add_notification(self, flow: FlowEvent, receiver_config: WatchConfig, raw_tx):
        pass

    @abstractmethod
    def send_notifications(self):
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class FlowProvider(ABC):
    @abstractmethod
    def get_flows_for_block(
        self, block: int
    ) -> Optional[List[Tuple[FlowEvent, object]]]:
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class WatcherState(ABC):
    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def get_next_watch_block(self) -> int:
        pass

    @abstractmethod
    def done_with_block(self):
        pass

    @abstractmethod
    def persist(self):
        pass

    def __enter__(self):
        self.load()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import logging
import sys
from datetime import date, timedelta

import click

from ..cli.common import require_currency, require_environment
from ..config import config, supported_fiat_currencies
from ..utils.console import console
from .coindesk import MIN_START as MS_CD
from .coindesk import fetch as fetchCD
from .coindesk import ingest as ingestCD
from .coingecko import fetch as fetchGecko
from .coingecko import fetch_impl as fetchGeckoDump
from .coingecko import ingest as ingestGecko
from .coinmarketcap import MIN_START as MS_CMK
from .coinmarketcap import fetch as fetchCMK
from .coinmarketcap import fetch_impl as fetchCMKDump
from .coinmarketcap import ingest as ingestCMK
from .cryptocompare import MIN_START as MS_CC
from .cryptocompare import fetch as fetchCC
from .cryptocompare import fetch_impl as dumpCC
from .cryptocompare import ingest as ingestCC

logger = logging.getLogger(__name__)


def get_api_key(key):
    api_key_key = f"{key}_api_key"
    api_key = getattr(config, api_key_key)
    if not api_key.strip():
        logger.error(f"Please provide an API key (graphsense.yaml -> {api_key_key})")
        sys.exit(1)
    return api_key


def shared_flags(provider="cmc"):
    def inner(function):
        prev_date = (date.today() - timedelta(days=1)).strftime("%Y-%m-%d")
        min_date = "2009-01-03"
        if provider == "cmc":
            min_date = MS_CMK
        elif provider == "cdesk":
            min_date = MS_CD
        elif provider == "cryptocompare":
            min_date = MS_CC

        function = click.option(
            "--fiat-currencies",
            default=supported_fiat_currencies,
            multiple=True,
            show_default=True,
            help="List of fiat currencies.",
        )(function)

        function = click.option(
            "--start-date",
            default=min_date,
            type=str,
            help="start date for fetching exchange rates",
        )(function)

        function = click.option(
            "--end-date",
            default=prev_date,
            show_default=True,
            type=str,
            help="end date for fetching exchange rates",
        )(function)

        return function

    return inner


def shared_ingest_flags():
    def inner(function):
        function = click.option(
            "--force",
            "-f",
            is_flag=True,
            help=(
                "don't continue from last found Cassandra record "
                "and force overwrite of existing rows"
            ),
        )(function)

        function = click.option(
            "-t",
            "--table",
            default="exchange_rates",
            show_default=True,
            type=str,
            help="Name of the target exchange rate table",
        )(function)

        function = click.option(
            "--abort-on-gaps",
            is_flag=True,
            help=(
                "ECB does not provide courses on weekends and holidays. "
                "The default behavior of the script is to fill (interpolate). "
                "With this flag the script aborts if gaps are still present."
            ),
        )(function)

        function = click.option(
            "--dry-run",
            is_flag=True,
            help="Don't write new records to Cassandra.",
        )(function)
        return function

    return inner


@click.group()
def rates_cli():
    pass


@rates_cli.group()
def exchange_rates():
    """Fetching and ingesting exchange rates."""
    pass


@exchange_rates.group()
def coindesk():
    """From coindesk."""
    pass


@exchange_rates.group()
def coinmarketcap():
    """From coinmarketcap."""
    pass


@exchange_rates.group()
def coingecko():
    """From coingecko."""
    pass


@exchange_rates.group()
def cryptocompare():
    """From cryptocompare."""
    pass


@coinmarketcap.command("dump")
@require_currency()
@shared_flags()
@click.option(
    "--out-file",
    default="rates.csv",
    type=str,
    help="file to dump into.",
)
def fetch_cmk_dump(
    currency: str,
    fiat_currencies: list[str],
    start_date: str,
    end_date: str,
    out_file: str,
):
    """Save exchange rates to file.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
    """
    api_key = get_api_key("coinmarketcap")
    df = fetchCMKDump(
        None,
        None,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        None,
        True,
        False,
        False,
        api_key,
    )
    console.rule("Rates Coinmarketcap")
    console.print(df)
    console.rule(f"Writing to {out_file}")
    df.to_csv(out_file)


@coingecko.command("dump")
@require_currency()
@shared_flags(provider="gecko")
@click.option(
    "--out-file",
    default="rates.csv",
    type=str,
    help="file to dump into.",
)
def fetch_coingecko_dump(
    currency: str,
    fiat_currencies: list[str],
    start_date: str,
    end_date: str,
    out_file: str,
):
    """Save exchange rates to file.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
    """
    api_key = get_api_key("coingecko")
    df = fetchGeckoDump(
        None,
        None,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        None,
        True,
        False,
        False,
        api_key,
    )
    console.rule("Rates Coingecko")
    console.print(df)
    console.rule(f"Writing to {out_file}")
    df.to_csv(out_file)


@cryptocompare.command("dump")
@require_currency()
@shared_flags(provider="cryptocompare")
@click.option(
    "--out-file",
    default="rates.csv",
    type=str,
    help="file to dump into.",
)
def fetch_cryptocompare_dump(
    currency: str,
    fiat_currencies: list[str],
    start_date: str,
    end_date: str,
    out_file: str,
):
    """Save exchange rates to file.
    \f
    Args:
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
        out_file (str): -
    """
    df = dumpCC(
        None,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        None,
        True,
        True,
        False,
    )
    console.rule("Rates Coingecko")
    console.print(df)
    console.rule(f"Writing to {out_file}")
    df.to_csv(out_file)


@coinmarketcap.command("fetch")
@require_environment()
@require_currency()
@shared_flags()
def fetch_cmk(
    env: str, currency: str, fiat_currencies: list[str], start_date: str, end_date: str
):
    """Only fetches the to be imported exchange rates.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
    """
    api_key = get_api_key("coinmarketcap")
    df = fetchCMK(env, currency, list(fiat_currencies), start_date, end_date, api_key)
    console.rule("Rates Coinmarketcap")
    console.print(df)


@coingecko.command("fetch")
@require_environment()
@require_currency()
@shared_flags(provider="gecko")
def fetch_gecko(
    env: str, currency: str, fiat_currencies: list[str], start_date: str, end_date: str
):
    """Only fetches the to be imported exchange rates.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
    """
    api_key = get_api_key("coingecko")
    df = fetchGecko(env, currency, list(fiat_currencies), start_date, end_date, api_key)
    console.rule("Rates Coingecko")
    console.print(df)


@coindesk.command("fetch")
@require_environment()
@require_currency()
@shared_flags(provider="cdesk")
def fetch_cd(
    env: str, currency: str, fiat_currencies: list[str], start_date: str, end_date: str
):
    """Only fetches the to be imported exchange rates.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
    """
    df = fetchCD(env, currency, list(fiat_currencies), start_date, end_date)
    console.rule("Rates Coindesk")
    console.print(df)


@cryptocompare.command("fetch")
@require_environment()
@require_currency()
@shared_flags(provider="cryptocompare")
def fetch_cc(
    env: str, currency: str, fiat_currencies: list[str], start_date: str, end_date: str
):
    """Fetches and prints exchange rates.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
    """
    df = fetchCC(env, currency, list(fiat_currencies), start_date, end_date)
    console.rule("Rates cryptocompare")
    console.print(df)


@coinmarketcap.command("ingest")
@require_environment()
@require_currency()
@shared_flags()
@shared_ingest_flags()
def ingest_cmk(
    env: str,
    currency: str,
    fiat_currencies: list[str],
    start_date: str,
    end_date: str,
    table: str,
    force: bool,
    dry_run: bool,
    abort_on_gaps: bool,
):
    """Ingest exchange rates into Cassandra
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
        table (str): -
        force (bool): -
        dry_run (bool): -
        abort_on_gaps (bool): -
    """
    api_key = get_api_key("coinmarketcap")
    ingestCMK(
        env,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        table,
        force,
        dry_run,
        abort_on_gaps,
        api_key,
    )


@coingecko.command("ingest")
@require_environment()
@require_currency()
@shared_flags()
@shared_ingest_flags()
def ingest_gecko(
    env: str,
    currency: str,
    fiat_currencies: list[str],
    start_date: str,
    end_date: str,
    table: str,
    force: bool,
    dry_run: bool,
    abort_on_gaps: bool,
):
    """Ingest exchange rates into Cassandra
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
        table (str): -
        force (bool): -
        dry_run (bool): -
        abort_on_gaps (bool): -
    """
    api_key = get_api_key("coingecko")
    ingestGecko(
        env,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        table,
        force,
        dry_run,
        abort_on_gaps,
        api_key,
    )


@coindesk.command("ingest")
@require_environment()
@require_currency()
@shared_flags(provider="cdesk")
@shared_ingest_flags()
def ingest_cd(
    env,
    currency,
    fiat_currencies,
    start_date,
    end_date,
    table,
    force,
    dry_run,
    abort_on_gaps,
):
    """Ingests new exchange rates into cassandra raw keyspace.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
        table (str): -
        force (bool): -
        dry_run (bool): -
        abort_on_gaps (bool): -
    """
    ingestCD(
        env,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        table,
        force,
        dry_run,
        abort_on_gaps,
    )


@cryptocompare.command("ingest")
@require_environment()
@require_currency()
@shared_flags(provider="cryptocompare")
@shared_ingest_flags()
def ingest_cc(
    env,
    currency,
    fiat_currencies,
    start_date,
    end_date,
    table,
    force,
    dry_run,
    abort_on_gaps,
):
    """Ingests new exchange rates into cassandra raw keyspace.
    \f
    Args:
        env (str): -
        currency (str): -
        fiat_currencies (list[str]): -
        start_date (str): -
        end_date (str): -
        table (str): -
        force (bool): -
        dry_run (bool): -
        abort_on_gaps (bool): -
    """
    ingestCC(
        env,
        currency,
        list(fiat_currencies),
        start_date,
        end_date,
        table,
        force,
        dry_run,
        abort_on_gaps,
    )



from typing import Optional

from pydantic import BaseModel


class BlockchainAdapter:
    datamodel = None
    name_remapping = {}
    field_processing = {}
    dataclass_name = ""

    def dict_to_dataclass(self, data_dict):
        return self.datamodel.parse_obj(data_dict)

    def process_fields(self, data_object):
        # Check if the object is an instance of a dataclass
        for field_name, field_processor in self.field_processing.items():
            setattr(
                data_object,
                field_name,
                field_processor(getattr(data_object, field_name)),
            )
        return data_object

    def rename_dict(self, data_dict):
        for old_name, new_name in self.name_remapping.items():
            if old_name in data_dict:
                data_dict[new_name] = data_dict.pop(old_name)
        return data_dict

    def dict_to_renamed_dataclass(self, data_dict):
        dc = self.datamodel
        renamed_dict = self.rename_dict(data_dict)
        data_req = {k: v for k, v in renamed_dict.items() if k in dc.__annotations__}
        return dc(**data_req)

    def dicts_to_dataclasses(self, data_dicts):
        return [self.dict_to_dataclass(data_dict) for data_dict in data_dicts]

    def dicts_to_renamed_dataclasses(self, data_dicts):
        return [self.dict_to_renamed_dataclass(data_dict) for data_dict in data_dicts]

    def process_fields_in_list(self, data_list):
        return [self.process_fields(data_object) for data_object in data_list]


class Trace(BaseModel):
    block_id: int
    tx_hash: Optional[bytes]
    trace_index: int
    from_address: Optional[bytes]
    to_address: Optional[bytes]
    value: int
    call_type: Optional[str]
    status: int


class Transaction(BaseModel):
    transaction_index: int
    tx_hash: bytes
    from_address: Optional[bytes]
    to_address: Optional[bytes]
    value: int
    gas_price: int
    transaction_type: int
    receipt_gas_used: int
    receipt_status: int
    block_id: int


class TronTransaction(Transaction):
    fee: Optional[int]


class Log(BaseModel):
    block_id: int
    tx_hash: bytes
    log_index: int
    address: bytes
    topics: list
    data: bytes


class Block(BaseModel):
    block_id: int
    miner: bytes
    base_fee_per_gas: Optional[int]
    gas_used: int


class AccountTraceAdapter(BlockchainAdapter):
    datamodel = Trace


class AccountTransactionAdapter(BlockchainAdapter):
    datamodel = Transaction


class TrxTransactionAdapter(BlockchainAdapter):
    datamodel = TronTransaction


class AccountLogAdapter(BlockchainAdapter):
    datamodel = Log


class TrxTraceAdapter(AccountTraceAdapter):
    def __init__(self):
        self.name_remapping = {
            "caller_address": "from_address",
            "transferto_address": "to_address",
            "rejected": "status",
            "note": "call_type",
            "call_value": "value",
        }
        self.field_processing = {"status": lambda x: int(not x)}  # cast boolean to int


class EthTraceAdapter(AccountTraceAdapter):
    pass


def zero_if_none(x):
    return x if x is not None else 0


class AccountBlockAdapter(BlockchainAdapter):
    datamodel = Block

    def __init__(self):
        self.field_processing = {"base_fee_per_gas": zero_if_none}
